import os

from typing import BinaryIO, cast
from contextlib import contextmanager

from .abstract import AbstractAdapter


class FilesystemAdapter(AbstractAdapter):
    def test_file(self, file_path: str) -> bool:
        return os.path.isfile(file_path) and os.access(file_path, os.R_OK)

    def read_file(self, file_name: str, size=-1, offset=0) -> bytes:
        with self.open(file_name) as file_h:
            file_h = cast(BinaryIO, file_h)
            file_h.seek(offset)
            return file_h.read(size)

    def get_img_bd_size(self) -> int:
        assert self.img_bd_path is not None
        return os.stat(self.img_bd_path).st_size

    def find_file(self, file_name: str) -> "str | None":
        all_files = next(os.walk(self.load_path))[2]
        file_name_upper = file_name.upper()
        file_match_gen = (file_name for file_name in all_files if file_name.upper() == file_name_upper)
        file_name_match = next(file_match_gen, None)

        if file_name_match is not None:
            file_name_match = os.path.join(self.load_path, file_name_match)

        return file_name_match

    @contextmanager
    def open(self, file_name: str) -> BinaryIO:  # pyright: ignore[reportGeneralTypeIssues]
        with open(file_name, "rb") as file_h:
            yield cast(BinaryIO, file_h)



import re


re_table = re.compile(r"({.*?}|.)")


def make_table(font):
    font = font.strip("\n").replace("\n", "")
    table = re_table.findall(font)
    for n, x in enumerate(table):
        if x == "#":
            table[n] = f"{{0x{n:02X}}}"

    assert len(table) == 21 * 10

    return table



import os
import io
import re
import math
import struct
import pycdlib
import hashlib
import itertools

from abc import ABC, abstractmethod
from typing import BinaryIO, cast
from functools import reduce

from ..pk2 import PK2Archive
from ..tim2 import patch_pl_mtop
from ...utils.file import SubFile
from ..text.parser import inject_english_subtitles
from ..reader.entry import TOCEntry
from ..reader.pjzreader import PJZReader
from ...wagrenier.pssmux import pss_mux_from_bytes_io


class AbstractUndubEntry(ABC):
    @property
    @abstractmethod
    def name(self):
        ...

    @property
    @abstractmethod
    def number(self):
        ...

    @property
    @abstractmethod
    def data_size(self):
        ...

    @property
    @abstractmethod
    def new_size(self):
        ...

    @abstractmethod
    def read(self, size=-1):
        ...

    @abstractmethod
    def seek(self, offset: int):
        ...

    def close(self):
        pass


class ReaderUndubEntry(AbstractUndubEntry):
    @property
    def name(self):
        return self._name

    @property
    def number(self):
        return self._number

    @property
    def data_size(self):
        return self._toc_entry.size

    @property
    def new_size(self):
        return self._new_size

    def __init__(self, reader: PJZReader, toc_entry: TOCEntry, name: str, number: int, new_size=0, japanese=False):
        self._reader: PJZReader = reader
        self._toc_entry: TOCEntry = toc_entry
        self._name: str = name
        self._number: int = number
        self.japanese: bool = japanese

        if new_size == 0:
            self._new_size = toc_entry.size
        elif new_size > 0:
            self._new_size = max(new_size, toc_entry.size)

        self._offset = 0

    def read(self, size=-1):
        return self._reader.read_file(self._toc_entry.name, size=size, offset=self._offset)

    def seek(self, offset: int):
        self._offset = offset


class ExternalFileEntry(AbstractUndubEntry):
    @property
    def name(self):
        return self._name

    @property
    def number(self):
        return self._number

    @property
    def data_size(self):
        return self._data_size

    @property
    def new_size(self):
        return self._new_size

    def __init__(self, file: "str | bytes | BinaryIO", name: str, number: int, new_size=0):
        if isinstance(file, io.IOBase):
            self._file_h = file
        elif isinstance(file, bytes):
            self._file_h = io.BytesIO(file)
        else:
            self._file_h = open(  # pylint: disable=consider-using-with
                file,  # pyright: ignore[reportGeneralTypeIssues]
                "rb",
            )

        self._name = name
        self._number = number
        self._data_size = self._get_file_size()

        if new_size == 0:
            self._new_size = self._data_size
        elif new_size > 0:
            self._new_size = max(new_size, self._data_size)

    def _get_file_size(self):
        pos = self._file_h.tell()
        size = self._file_h.seek(0, os.SEEK_END)
        self._file_h.seek(pos, os.SEEK_SET)
        return size

    def close(self):
        self._file_h.close()

    def read(self, size=-1):
        return self._file_h.read(size)

    def seek(self, offset: int):
        return self._file_h.seek(offset, os.SEEK_SET)


def recalculate_img_bin_offsets(all_sizes, align=16):
    offsets = [0]
    for size in all_sizes:
        current_offset = offsets[-1]
        next_offset = current_offset + int(math.ceil(size / align / 2048)) * align
        offsets.append(next_offset)

    offset, img_bd_size = offsets[:-1], offsets[-1] * 2048

    return offset, img_bd_size


def compute_elf_crc(file_h: BinaryIO):
    pos = file_h.tell()
    file_h.seek(0, os.SEEK_SET)
    elf = file_h.read()
    u32s = struct.unpack(f"<{len(elf) // 4}I", elf)
    crc = reduce(lambda x, y: x ^ y, u32s, 0x00000000)
    file_h.seek(pos, os.SEEK_SET)
    return crc


def restore_elf_crc(file_h: BinaryIO, target: int):
    pos = file_h.tell()
    crc = compute_elf_crc(file_h)
    crc_fix = struct.pack("<I", crc ^ target)
    file_h.seek(0x08, os.SEEK_SET)
    file_h.write(crc_fix)
    file_h.seek(pos, os.SEEK_SET)


def patch_elf_inplace(
    iso_path: str,
    fix_kirie_camera_bug=True,
    force_lang=False,
    no_bloom=False,
    dark_filter=False,
    ingame_noise=False,
    menu_noise=False,
    force_16_9_game=False,
    force_16_9_movies=False,
    callback=None,
):
    iso = pycdlib.PyCdlib()
    iso.open(iso_path, mode="rb")

    record = iso.get_record(iso_path="/SLES_508.21;1")
    offset = record.fp_offset
    size = record.data_length

    iso.close()

    if callback:
        callback(
            2
            + int(fix_kirie_camera_bug)
            + int(force_lang)
            + int(no_bloom)
            + int(dark_filter)
            + int(ingame_noise)
            + int(menu_noise)
            + int(force_16_9_game)
            + int(force_16_9_movies)
        )

    with open(iso_path, mode="rb+") as iso_fh:
        file_h = SubFile(iso_fh, offset, size)  # pylint: disable=abstract-class-instantiated

        original_crc = compute_elf_crc(file_h)

        # enable english subtitles
        file_h.seek(0x0005691A)
        file_h.write(b"\x00\x14")
        file_h.seek(0x00056952)
        file_h.write(b"\x00\x14")
        file_h.seek(0x00056B12)
        file_h.write(b"\x00\x10")
        file_h.seek(0x000613B2)
        file_h.write(b"\x00\x14")

        if callback:
            callback()

        if force_lang:
            file_h.seek(0x001202CE)
            file_h.write(b"\x00\x14")

            if callback:
                callback()

        if fix_kirie_camera_bug:
            file_h.seek(0x000203B4)
            file_h.write(b"\x32\x60\x15\x46\x02\x00\x01\x45")

            if callback:
                callback()

        if no_bloom:
            file_h.seek(0x00251C0E)
            file_h.write(b"\x00\x00")

            if callback:
                callback()

        if dark_filter:
            file_h.seek(0x0025208E)
            file_h.write(b"\x00\x00")

            if callback:
                callback()

        if ingame_noise:
            file_h.seek(0x00251F1E)
            file_h.write(b"\x00\x00")

            if callback:
                callback()

        if menu_noise:
            file_h.seek(0x0025A05E)
            file_h.write(b"\x00\x00")

            if callback:
                callback()

        if force_16_9_game:
            file_h.seek(0x00036B18)
            file_h.write(b"\x8C")
            file_h.seek(0x00036B80)
            file_h.write(b"\xA8")
            file_h.seek(0x00036BC4)
            file_h.write(b"\x28")
            file_h.seek(0x00036BFC)
            file_h.write(b"\x0C")
            file_h.seek(0x0003815C)
            file_h.write(b"\x12")
            file_h.seek(0x00086B40)
            file_h.write(b"\xC0")
            file_h.seek(0x00086B4C)
            file_h.write(b"\x40")
            file_h.seek(0x0008B2CC)
            file_h.write(b"\x40")

            if callback:
                callback()

        if force_16_9_movies:
            file_h.seek(0x00083731)
            file_h.write(b"\x71")
            file_h.seek(0x00083741)
            file_h.write(b"\x71")
            file_h.seek(0x00083749)
            file_h.write(b"\x1E")

            if callback:
                callback()

        restore_elf_crc(file_h, target=original_crc)

        if callback:
            callback()


def patch_english_subtitles(reader: PJZReader, ig_msg_entry: TOCEntry):
    with reader.open(ig_msg_entry.name) as file_h:
        patched_ig_msg = inject_english_subtitles(file_h)
        return ExternalFileEntry(patched_ig_msg, ig_msg_entry.name, ig_msg_entry.number)


def replace_movies_in_iso_inplace(iso_jp_path, iso_undub_path, callback=None):
    iso_jp = pycdlib.PyCdlib()
    iso_jp.open(iso_jp_path, mode="rb")

    iso_undub = pycdlib.PyCdlib()
    iso_undub.open(iso_undub_path, mode="rb")

    jp_video_movie = [f"/MOVIE/{f}" for f in next(iso_jp.walk(iso_path="/MOVIE"))[2] if ".PSS;" in f]
    jp_video_movie2 = [f"/MOVIE2/{f}" for f in next(iso_jp.walk(iso_path="/MOVIE2"))[2] if ".PSS;" in f]

    en_video_movie = [f"/MOVIE/{f}" for f in next(iso_undub.walk(iso_path="/MOVIE"))[2] if ".PSS;" in f]
    en_video_movie2 = [f"/MOVIE2/{f}" for f in next(iso_undub.walk(iso_path="/MOVIE2"))[2] if ".PSS;" in f]
    en_video_movie3 = [f"/MOVIE3/{f}" for f in next(iso_undub.walk(iso_path="/MOVIE3"))[2] if ".PSS;" in f]
    en_video_movie4 = [f"/MOVIE4/{f}" for f in next(iso_undub.walk(iso_path="/MOVIE4"))[2] if ".PSS;" in f]

    # en_video_movie5 = [f'/MOVIE5/{f}' for f in next(iso_undub.walk(iso_path='/MOVIE5'))[2] if '.PSS;' in f]

    def find_matches(_jp_list, _en_list):
        def basename(p):
            return os.path.splitext(os.path.basename(p))[0].rstrip("P")

        matches = list()
        for _jp_movie in _jp_list:
            en_match = next((_en_movie for _en_movie in _en_list if basename(_en_movie) == basename(_jp_movie)), None)
            if en_match:
                matches.append((_jp_movie, en_match))

        return matches

    matches_movie = find_matches(jp_video_movie, en_video_movie)
    matches_movie_p = find_matches(jp_video_movie, en_video_movie3)

    matches_movie2 = find_matches(jp_video_movie2, en_video_movie2)
    matches_movie2_p = find_matches(jp_video_movie2, en_video_movie4)

    all_matches = matches_movie + matches_movie_p + matches_movie2 + matches_movie2_p

    if callback:
        callback(len(all_matches))

    for jp_movie, en_movie in all_matches:
        record_undub = iso_undub.get_record(iso_path=en_movie)
        offset_undub = record_undub.fp_offset
        size_undub = record_undub.data_length

        with iso_jp.open_file_from_iso(iso_path=jp_movie) as jp_movie_fh:
            with open(iso_undub_path, "rb+") as iso_undub_fh:
                en_movie_fh = SubFile(  # pylint: disable=abstract-class-instantiated
                    iso_undub_fh, offset=offset_undub, size=size_undub
                )
                pss_mux_from_bytes_io(jp_movie_fh, en_movie_fh)  # pyright: ignore[reportGeneralTypeIssues]

        if callback:
            callback()


def compute_file_hash(iso_path: str, file_name: str):
    iso = pycdlib.PyCdlib()
    iso.open(iso_path, mode="rb")

    with iso.open_file_from_iso(iso_path=f"/{file_name.upper()};1") as f:
        sha256_hash = hashlib.sha256()
        for byte_block in iter(lambda: f.read(16 * 1024), b""):
            sha256_hash.update(byte_block)

    return sha256_hash.hexdigest()


def check_iso_hashes(iso_file: str, lang: str, callback=None):
    known_hashes_eu = {
        "SLES_508.21": "cb7c8b5552c245ec17b28a26347e1801c9e90eee4a99c7b3d86bbdea088fbb3a",
        "IMG_HD.BIN": "ab3c26678d42458405e753715d820df766e332a5027f43b2f9e1c8901d0e0657",
        "IMG_BD.BIN": "f18aeff01761e539aea41f830a94d0468c80a2a9cf8bb700e51aba0752fe8bca",
        "MOVIE/DUMMY.PSS": "010cc1efb67bbf21491b3a53043ffeef71f8fea4d2450393488ba1cd67a34775",
        "MOVIE/SCN0010.PSS": "3961be41d6a14f27622c2b21e5bed57f3c8238d0fdefe402a8bf917c9cffe702",
        "MOVIE/SCN0031.PSS": "c2ccb7e5a9c1d8226fb642d4f89b64ac5d0ca92cf07649af7697323281b1833e",
        "MOVIE/SCN1000.PSS": "dcb5638cd634b11af724bef2a11e11e0518ddebf5d211a9b9b8504072ac9aa13",
        "MOVIE/SCN1031.PSS": "17e30f7795f8c97056924268e011fa07d2a66f82252f7486ec1bb4a02579316a",
        "MOVIE/SCN1101.PSS": "02359a28a2bbec031723bbe80d37aa89329c601da0f5f112b751d4b67f1e3680",
        "MOVIE/SCN1240.PSS": "36e809797659ab961e41058c924fb852407913c25ce873acc69a3d6e6dc4a332",
        "MOVIE/SCN1300.PSS": "d01f44847d12ceea635492a81ece713ab28947137d9a5764f359d7dd83b1198c",
        "MOVIE/SCN1331.PSS": "c5fdde597f64bae03870f9d08a46a7b4d28814c739fca07ecbd6d17ad4786643",
        "MOVIE/SCN1332.PSS": "6533f79b4bd2056bea8eb2d7969c27ed8919231cf13160b4afd91a9e41a5adab",
        "MOVIE/SCN2010.PSS": "d254cd8b0b5da2e6442d6e14d8fdf613e00354366d49691ddddf28f62702b235",
        "MOVIE/SCN2050.PSS": "b548d6678ceb453755adc5deaaa5a7a631bdc280bb8190fabebdbd034ab221b6",
        "MOVIE/SCN2061.PSS": "3cb9b77baaa3370283fc42c5d5a3897086c50366f9d015184cd654c300b26e38",
        "MOVIE/SCN2071.PSS": "5a4d8d33ef7b1df63a1132e1e7057d995c1144bccef3ba1feaa5d8f81284cdb4",
        "MOVIE/SCN2091.PSS": "cff8b9fad2f4b2648cb53b4671960bd4031f561512c048d05e7e5aaf8015bc06",
        "MOVIE/SCN2110.PSS": "0f58f0b1b8948ad30dc35768f8ca966febeeec50b653d3052f963532fae76e74",
        "MOVIE/SCN2131.PSS": "35814d1e8b617e3a1fc9f08ce25af7ac94bb3fa5fc25bb17ea91c9c6a694bc64",
        "MOVIE/SCN2142.PSS": "675bfe9f6cd8236c74c00a2a9ec5c420126b519285e971e1b45e6b8b7ddc5673",
        "MOVIE/SCN2143.PSS": "2cb68bec2ece0bc64e3a976865cb5d93fc757d52ccf3226c6f1923d20afdab14",
        "MOVIE/SCN2171.PSS": "d730b85838be4eb2077566eec1a3a5d6ad0d92b5b666570679c2a88aa171edd3",
        "MOVIE/SCN9000.PSS": "ea18c9cbf9ecb3b320926e5cf57020032fe005e9c6b8bf8ce3d6a5d3d1ef18a6",
        "MOVIE/SCN9001.PSS": "ebf2220545675640c0a51161bdd748677f571385b41992704a497594f653d4e8",
        "MOVIE/SCN9100.PSS": "5e75d4f60c61b69185f496b35753b0eb5f77c07720cdcc89e7049e69a3161d03",
        "MOVIE/SCN9200.PSS": "990c35a26ff1fe5fe237f28b98d5d132dd4c52847dfee6cd789734f8980f2b65",
        "MOVIE/TECMO.PSS": "c5f6aee96d9fb8729cfd65a44fd56548001613e9a8a7c0d25f89c7b5e1b078ec",
        "MOVIE2/SCN3010.PSS": "06bb0a76f5c6e894287d67d6c6ef805c338a1a4dceda76f4b30616020651a68f",
        "MOVIE2/SCN3040.PSS": "54762d5275d873e90398cbbf0c48113965c2830ad7e98341857545d3d6efe37e",
        "MOVIE2/SCN3080.PSS": "373e013ed60a604f59e290f376989783e7e7198f067f48bf15258450cdb3083e",
        "MOVIE2/SCN3081.PSS": "778405b1aa23c1b3e04b14b7d9b0bb79a2036b35cbcf57d193024110b61602ed",
        "MOVIE2/SCN3090.PSS": "82bcb89ca31c9ad28d6ff6e336986969807374447eb9629e9ba2cbcac1e68281",
        "MOVIE2/SCN4010.PSS": "55f31b6ec122a63c28694a7f210f930632edf1fc1f76f0682e218626efcd09f2",
        "MOVIE2/SCN4031.PSS": "1f5ef156e1bef2092d828ad671e8df9f86f0b46a9d9b010cd309aa123e526428",
        "MOVIE2/SCN4041.PSS": "8b859a2844db2826bb68aae453ec2cdc5847d51c93dbf0392367cabd66b5b66d",
        "MOVIE2/SCN4060.PSS": "d83fb519c5d8f26f1877f081535974bc7352d25ec445f1e73bf306ae2095b9d4",
        "MOVIE2/SCN4080.PSS": "94e06f3d1cddd975ab6711b79565a7ab6267f0fddd0ffb7e688b3b09f59345bd",
        "MOVIE2/SCN4090.PSS": "b162b0fd61231de7ff3cfbfd89436969e2058dcc90022c1bf916ab88be2283b5",
        "MOVIE2/SCN4100.PSS": "00c8883983235c2bea06cda23385da2c6ade3f9a8959e78a19469002ef75628d",
        "MOVIE2/SCN4110.PSS": "32e80c5cd3e21e96fc2ea6b8daf5f2f3fa0ae42818cf2015b0ca7c347932a4d8",
        "MOVIE2/SCN4120.PSS": "d00000f98514cd9d65ce0d5ae627235f0d0b658480e489d870642eb6ac283c09",
        "MOVIE2/SCN5010.PSS": "2403c40b1e9619e9eb0d1c4f1d56ef82aeb42d566f2ef51cff7b0b72b837fac5",
        "MOVIE2/SCN5020.PSS": "e1319796e3757a31a55d3012d54dc2d7d04d6410a1f2471b9de8c5c48fc03469",
        "MOVIE3/SCN0010P.PSS": "f8b163ed827afff9aefb8d4b25651bfd4cfcef7cc4757bc9ce26314ba30f5cb4",
        "MOVIE3/SCN0031P.PSS": "ed191c4fb9392e413681dce24f26edd14cbe91943f6d10fe2182386f299b5a92",
        "MOVIE3/SCN1000P.PSS": "8871f436fe1fa9073272cee28a095bee951411fb3afb7b49da5a5b99322078bb",
        "MOVIE3/SCN1031P.PSS": "27871f71baa934450f118e0ebeb93f4914c418f687b5d5a7c08684d0ec976c97",
        "MOVIE3/SCN1101P.PSS": "06719a2db5997c67f251cba62d7932e1ae2aa1ef574b92a634515c6fa6deeeaf",
        "MOVIE3/SCN1240P.PSS": "04995368626b01307f028ddf2dc8fc0e37e1a98817ddde4436b13748f83354f1",
        "MOVIE3/SCN1300P.PSS": "260081eff01df63fe083df2435013a46bc0d3587810257f6472c83b20d00200d",
        "MOVIE3/SCN1331P.PSS": "39f8af7fbc81800bc1502616e66e1e58a979dd9d02db0bf3331d3b1ea9bb3f4b",
        "MOVIE3/SCN1332P.PSS": "bcba19a5cfceb7bbba264eaa7c0e542fb939c8d41cd70f60038a343495750742",
        "MOVIE3/SCN2010P.PSS": "bb0ac5e3d3d6323e53c1877fbfa11e73a496ae12acb00d74ce54f94e25a12f30",
        "MOVIE3/SCN2050P.PSS": "f44172e4c167f75b86dbd66ce83ec6a16935261dbc27ad5b028f6057f014a158",
        "MOVIE3/SCN2061P.PSS": "c873fa8271dd51490eddfda4ab58d64e3764f5d4c01f085e4b9cde507aac1d42",
        "MOVIE3/SCN2071P.PSS": "9bb1f424e6333044296f866c41e18c73bf18e5099674bfeda825d17e53e8f475",
        "MOVIE3/SCN2091P.PSS": "e4a35bcf27c2a55b572b28a28c3008228f9b73393ecc05c6507787a3b66e4285",
        "MOVIE3/SCN2110P.PSS": "39294e3c2147b033c034e8f579311048fe0fa72f00bf83ba04537ad463b3c5ea",
        "MOVIE3/SCN2131P.PSS": "5aff6ffb03f7a288e720b0168278c33fd1ec2556796b597d6d47867c63439836",
        "MOVIE3/SCN2142P.PSS": "0c673d6629b9460149e789872615702e5dfd0fd25a9c27207b1de4a3b72aab8e",
        "MOVIE3/SCN2143P.PSS": "36c530a33e224223ee808e8d983c34800ef6a4f5993996b016ef38715405a86f",
        "MOVIE3/SCN2171P.PSS": "4332e33ac86798e5cdeb41c93b13d62f3413a2a5c830c1ed61ea80f3ee7ec15f",
        "MOVIE3/SCN9000P.PSS": "2862d253ad97cb744199e65e04dd16131d2bd7e03aa758b6b96392304f894267",
        "MOVIE3/SCN9001P.PSS": "ebf2220545675640c0a51161bdd748677f571385b41992704a497594f653d4e8",
        "MOVIE3/SCN9100P.PSS": "b5313245d245cb7fe7d4822eccd3b1730789975a5fa750d925c2629145872c20",
        "MOVIE3/SCN9200P.PSS": "978210c7ceb4b271498a83b8bfe7ed912e72c78fcb64c5df23063a69afd10b22",
        "MOVIE3/TECMOP.PSS": "c5f6aee96d9fb8729cfd65a44fd56548001613e9a8a7c0d25f89c7b5e1b078ec",
        "MOVIE4/SCN3010P.PSS": "ec148a5ecb562d312652a039d0632f1c39aef7b89cd1a208ed0e0e0065000968",
        "MOVIE4/SCN3040P.PSS": "ca0928f5b6b2ef759e909ece89656b7c166831ccdb54fe84147c2f38665fd339",
        "MOVIE4/SCN3080P.PSS": "86325fa8b50f782cc7fe367e9d891d8c41d47d4de34e1c8d95048c86ad915e3c",
        "MOVIE4/SCN3081P.PSS": "2dace3583c771adccb4e6fcf338baec49b8dab08bb1a13c3ad6f3c6b780605e6",
        "MOVIE4/SCN3090P.PSS": "2d8e7d6fc06bd6f0a993463554ca7248341b1a4d8630b3d3921571d4957124cb",
        "MOVIE4/SCN4010P.PSS": "5473b20e920c47a2e5a364872859043360c4be432da66b0c9da76b7ea0f46504",
        "MOVIE4/SCN4031P.PSS": "c6f853f8751837f120b55dc14e23b2ec2aac4b69e3a5b2698875fff7e7cee065",
        "MOVIE4/SCN4041P.PSS": "17ccb610d914b41305b0b93f0746b2175396cc830bb63d2739b3e638532964d6",
        "MOVIE4/SCN4060P.PSS": "ac9ba38399df64bf57a915fe65beb288cb865502993e39f16321f6bcc91523e6",
        "MOVIE4/SCN4080P.PSS": "da5ae8364ee38048f62bdd716a422e38980936a0397bd804609a810acfbd27cd",
        "MOVIE4/SCN4090P.PSS": "ea609e551421d47dc17a56d9d3df72247a75d794c7037bbd0b57d315f5c56dd0",
        "MOVIE4/SCN4100P.PSS": "f6d90cee4348e74c547bd2ca2f9ca1fdca4c12c11a923044e25cac4949a28314",
        "MOVIE4/SCN4110P.PSS": "6099093b39044aea2fd6ec728600ff073c08f684f8078881fa3f6001982e7bd9",
        "MOVIE4/SCN4120P.PSS": "367e807d6435c2f3a1231c2d0eb5002021db626611e8911e9132fa99aedf2849",
        "MOVIE4/SCN5010P.PSS": "58c3c4af2c34bc5d5bb005ef2d93604e6da6e60909db0ecf578fa6a38128527e",
        "MOVIE4/SCN5020P.PSS": "fe4b2a0b210f9f5197d3b7151313bae2ce4c6d4b7326b6a46d73cb682986319b",
        "MOVIE5/SCN900EP.PSS": "e0f9317e5102c3951ec2f2ccdeaa35dd2005a4f5bbba17244a9a28be43725545",
        "MOVIE5/SCN900E.PSS": "4a2cfd688f51411a1bb23d2e3571dce4bc70935dda15f60d7811df21dfcdc703",
        "MOVIE5/SCN900FP.PSS": "b18c01a03cddcdfdb96d228ac673a5285c8aea6167d623a2d9e9d8d623686758",
        "MOVIE5/SCN900F.PSS": "b162c72674111152ecc7c331363869668d6da0d54473eda63f2a88f831f8d3b6",
        "MOVIE5/SCN900GP.PSS": "d7ef0d6f5a27e6b51fab8dd175b809bffed5f5d8304e8f9f6922ec0f1dd17410",
        "MOVIE5/SCN900G.PSS": "0256e8ecff005d95cf149ab97c9c0fe6c85e5f709f7fe1515ec4f547b0486f00",
        "MOVIE5/SCN900IP.PSS": "b0f8fb270a119c68e37e9958f4773a72e0fd81a8fed3e35e66ef7de549edffc1",
        "MOVIE5/SCN900I.PSS": "24e2d1d297d2ec2e1e11bb1d975a442fef7f6019bd1aa2fbb0d65d82797ac2d0",
        "MOVIE5/SCN900SP.PSS": "1d260910bc3476842edfeeb319e9409979a6707849c2dabdae59b34581c30d24",
        "MOVIE5/SCN900S.PSS": "f6115b89eeb200c637739491aa51b533d0523d5c008280b300cfbd5149b174e9",
    }
    known_hashes_jp = {
        "SLPS_250.74": "feb283ac7d09cc2f06275890885f3538c6267cdafd3f82249b8a44fed1fb8005",
        "IMG_HD.BIN": "9c8f270e15a78707251ea641bac8199d5f792e682e53e5e3400c71ab5e6a14fc",
        "IMG_BD.BIN": "e90cbc48f6f977acbbffc6107618ce1b86db03d08a54cefa9f52d903a9fa065d",
        "MOVIE/DUMMY.PSS": "010cc1efb67bbf21491b3a53043ffeef71f8fea4d2450393488ba1cd67a34775",
        "MOVIE/SCN0010.PSS": "f5c08c998f1d4b1c06a1756c60226a05cd654e926d51b32b0dccbc2f8565f6c3",
        "MOVIE/SCN0031.PSS": "3e537cc0b52d1c9ba2df384d2e34243fa9a80067f8db78518653af0c6ba1aeb3",
        "MOVIE/SCN1000.PSS": "a2d17acea663ef455c4b03214cd7056973ab09e00bf4c6345ada83f47536c212",
        "MOVIE/SCN1031.PSS": "a5385e9606e726a168f9f453fb3528231ed8507a50802edbddd1954406de8750",
        "MOVIE/SCN1101.PSS": "02359a28a2bbec031723bbe80d37aa89329c601da0f5f112b751d4b67f1e3680",
        "MOVIE/SCN1240.PSS": "b83ff53710719de785f695835fa5ea792412b60366155c1f687e8459aa4adc88",
        "MOVIE/SCN1300.PSS": "c07fa58905beb326e25001a3ac5b1ba745010b32f1e930839a976514c40f733a",
        "MOVIE/SCN1331.PSS": "c5fdde597f64bae03870f9d08a46a7b4d28814c739fca07ecbd6d17ad4786643",
        "MOVIE/SCN1332.PSS": "f2f0d3a1811089befd9b7cbec730fd842e436c542a39d4894ffeacaf55e99864",
        "MOVIE/SCN2010.PSS": "44a476e0a4d3370b67ee6c0a5de31ca2515c13bd58dae9c0068b290d228a8cb2",
        "MOVIE/SCN2050.PSS": "7a7fe5c45141999a151b46aa64daa102c99fee5fc73a180d70371b8bb7f6f29e",
        "MOVIE/SCN2061.PSS": "6946bb31edb03bdf72a7971b6540128745d0bf3f66bad6b19a376a978e5b5484",
        "MOVIE/SCN2071.PSS": "6f75e8942d29fceaa2166a8fe87f5eec1848e9db547b039396bd45d105f812ac",
        "MOVIE/SCN2091.PSS": "b5f0d34547881002f681b8a03ad029f9d3382e4d49714619c1ecc73935bde794",
        "MOVIE/SCN2110.PSS": "486aa3c8ccf73506216404d0e726e09bf97a74b7cd4429e9e825d5c2328ee094",
        "MOVIE/SCN2131.PSS": "02181cf19ac6617daa45faba5350ce7515f6da808ef11c40cdfcd22b65514193",
        "MOVIE/SCN2142.PSS": "4a1d8f5f57d816d71fb7a8b721f7c02da752777177b6dc0bb864f3ce852a886a",
        "MOVIE/SCN2143.PSS": "2cb68bec2ece0bc64e3a976865cb5d93fc757d52ccf3226c6f1923d20afdab14",
        "MOVIE/SCN2171.PSS": "5cb6fedb5cc43b4bfd22a71b3c4ec447ad7a8cc9dfe29f4efaa9ee94dc0af8b4",
        "MOVIE/SCN9000.PSS": "d1b5bbe6eb591f56b30fa80805d1dd9663ab1c5fcc76515ac3652f80604ba4db",
        "MOVIE/SCN9001.PSS": "120c7f8562e27e3c08f0159e54e05d0a91f5ab45937ce1cd54913b760dd214c4",
        "MOVIE/SCN9100.PSS": "f6972ea47cb6288a3e8021918f613361cd5e719da3648d9108440dce6d8baad4",
        "MOVIE/SCN9200.PSS": "bf29120de363ddc5e57c600736cf069e9c8fb89ec34fc4ba50531ad362668c4e",
        "MOVIE/TECMO.PSS": "c5f6aee96d9fb8729cfd65a44fd56548001613e9a8a7c0d25f89c7b5e1b078ec",
        "MOVIE2/SCN3010.PSS": "1523b503e3689c3e0375f30a3bec2c210a256f75d2a260c506958dbb529fbcc9",
        "MOVIE2/SCN3040.PSS": "3cabd74d2b4f79585ea88fe8dd89457afac8f09d46cc0ee0a9209e3270b35447",
        "MOVIE2/SCN3080.PSS": "3e4354dc7a94a6f47dc77391301bae434cea3933e865417e3d55b20205b830cb",
        "MOVIE2/SCN3081.PSS": "778405b1aa23c1b3e04b14b7d9b0bb79a2036b35cbcf57d193024110b61602ed",
        "MOVIE2/SCN3090.PSS": "82bcb89ca31c9ad28d6ff6e336986969807374447eb9629e9ba2cbcac1e68281",
        "MOVIE2/SCN4010.PSS": "8cb0c25f7c9a8eb9df93d7b93bf9bb20bdc7412d3a7d2b5f9eb07d1b3b25c381",
        "MOVIE2/SCN4031.PSS": "cfe1ff3f6b0e860ce512cc6b450578ebec741f4f5ffde516cae7b934361b45c3",
        "MOVIE2/SCN4041.PSS": "8b859a2844db2826bb68aae453ec2cdc5847d51c93dbf0392367cabd66b5b66d",
        "MOVIE2/SCN4060.PSS": "d83fb519c5d8f26f1877f081535974bc7352d25ec445f1e73bf306ae2095b9d4",
        "MOVIE2/SCN4080.PSS": "c0a6784fe206a4ea45c02cdecb568017e3d84d04e048d6f67dee81be6d48d6f0",
        "MOVIE2/SCN4090.PSS": "ca9f126e4d834275fe5996c79a9032208ccbce6fb518908d329586cf095be821",
        "MOVIE2/SCN4100.PSS": "00c8883983235c2bea06cda23385da2c6ade3f9a8959e78a19469002ef75628d",
        "MOVIE2/SCN4110.PSS": "0f932713a51e724b9eb9263e72e0fcccaee6927a4d41bf789e7668909a965daf",
        "MOVIE2/SCN4120.PSS": "1bbc5312dce70306e0bf5e1aa77995aef77f1d704cc0b45aa15f59c8297ac7ac",
        "MOVIE2/SCN5010.PSS": "fcbae8cbad3ac085299fce60487d959dd046a649d126dd4514fefa84b618eda5",
        "MOVIE2/SCN5020.PSS": "74a9e9e5574f1a5fcd62a6a86f5bb5420590e0fa3a5ef53ed9226af4ad6cb798",
    }

    if lang.upper() == "EU":
        known_hashes = known_hashes_eu
    elif lang.upper() == "JP":
        known_hashes = known_hashes_jp
    else:
        raise ValueError("lang must be EU or JP")

    if callback:
        callback(len(known_hashes))

    for file_name, hash_good in known_hashes.items():
        digest = compute_file_hash(iso_file, file_name)
        good = digest == hash_good

        if not good:
            raise RuntimeError

        if callback:
            callback()


def merge_iso_img_bd_contents(
    eu_iso_path: str,
    jp_iso_path: str,
    out_iso_path: str,
    replace_title_jp,
    replace_models,
    replace_sfx,
    callback=None,
):
    reader_eu = PJZReader(eu_iso_path)
    reader_jp = PJZReader(jp_iso_path)

    entries_eu_: list["TOCEntry | None"] = [reader_eu.find_entry(name) for name in reader_eu.list_files()]
    entries_jp_: list["TOCEntry | None"] = [reader_jp.find_entry(name) for name in reader_jp.list_files()]

    if not all(entries_eu_) or not all(entries_jp_):
        raise RuntimeError("cannot find all entries in ISO(s)")

    entries_eu = cast(list[TOCEntry], entries_eu_)
    entries_jp = cast(list[TOCEntry], entries_jp_)

    file_name_list_eu = [toc.name for toc in entries_eu]

    def filter_jp_entries_common(_pattern, pad=False):
        _eujp_entries: dict[str, ReaderUndubEntry] = {}
        for toc_jp in entries_jp:
            if re.match(_pattern, toc_jp.name) and toc_jp.name in file_name_list_eu:
                toc_eu = entries_eu[file_name_list_eu.index(toc_jp.name)]
                size = 0 if not pad else toc_eu.size
                _eujp_entries[toc_eu.name] = ReaderUndubEntry(
                    reader_jp, toc_jp, toc_eu.name, toc_eu.number, new_size=size, japanese=True
                )
        return _eujp_entries

    def filter_ingame_text_en():
        return next(_toc for _toc in entries_eu if _toc.name == "IG_MSG_E.OBJ")

    def repack_title():
        title_jp_toc = next((toc_jp for toc_jp in entries_jp if toc_jp.name == "TITLE.PK2"), None)
        if not title_jp_toc:
            raise RuntimeError("cannot find title image in japanese iso")

        titles_eu_toc = [toc_eu for toc_eu in entries_eu if re.match(r"TITLE_[EFGSI]\.PK2", toc_eu.name)]
        if len(titles_eu_toc) != 5:
            raise RuntimeError("cannot find title images in european iso")

        NUM_TIM2_IN_TITLE = 11

        new_titles: dict[str, ExternalFileEntry] = {}

        with reader_jp.open(title_jp_toc.name) as fh:
            title_jp_toc = PK2Archive(fh)

            for title_eu_toc in titles_eu_toc:
                with reader_eu.open(title_eu_toc.name) as fh:
                    title_eu = PK2Archive(fh, copy=True)

                    for i in range(NUM_TIM2_IN_TITLE):
                        title_eu[i] = title_jp_toc[i]

                    new_titles[title_eu_toc.name] = ExternalFileEntry(
                        file=title_eu.data,
                        name=title_eu_toc.name,
                        number=title_eu_toc.number,
                    )

        return new_titles

    def repack_pl_mtop():
        pl_mtop_jp = next((toc_jp for toc_jp in entries_jp if toc_jp.name == "PL_MTOP.PK2"), None)
        if not pl_mtop_jp:
            raise RuntimeError("cannot find pl_mtop image in japanese iso")

        pl_mtops_eu = [toc_eu for toc_eu in entries_eu if re.match(r"PL_MTOP_[EFGSI]\.PK2", toc_eu.name)]
        if len(pl_mtops_eu) != 5:
            raise RuntimeError("cannot find pl_mtop images in european iso")

        new_pl_mtop_eu: dict[str, ExternalFileEntry] = {}

        with reader_jp.open(pl_mtop_jp.name) as fh:
            archive = PK2Archive(fh)
            pl_mtop_jp_data_io = archive[1]

            for pl_mtop_eu in pl_mtops_eu:
                with reader_eu.open(pl_mtop_eu.name) as fh:
                    archive = PK2Archive(fh, copy=True)
                    pl_mtop_eu_data_io = archive[1]
                    archive[1] = patch_pl_mtop(pl_mtop_eu_data_io, pl_mtop_jp_data_io)

                    new_pl_mtop_eu[pl_mtop_eu.name] = ExternalFileEntry(
                        file=archive.data,
                        name=pl_mtop_eu.name,
                        number=pl_mtop_eu.number,
                    )

        return new_pl_mtop_eu

    def replace_models_untouched():
        untouched_model_names = (
            "M000_MIKU.MDL",
            "M000_MIKU.MPK",
            "M000_MIKU.PK2",
            "M000_SPE1.PK2",
            "M000_SPE2.PK2",
            "M000_SPE3.PK2",
            "REL11_MIKU.TM2",
            "TX_BTL_RES.PK2",
        )

        models_jp = {toc_jp.name: toc_jp for toc_jp in entries_jp if toc_jp.name in untouched_model_names}
        if len(models_jp) != len(untouched_model_names):
            raise RuntimeError("cannot find all M000 models in japanese iso")

        models_eu = {toc_eu.name: toc_eu for toc_eu in entries_eu if toc_eu.name in untouched_model_names}
        if len(models_eu) != len(untouched_model_names):
            raise RuntimeError("cannot find all M000 models in european iso")

        untouched_jp_entries: dict[str, ReaderUndubEntry] = {}

        for model_name in untouched_model_names:
            model_jp = models_jp[model_name]
            model_eu = models_eu[model_name]
            untouched_jp_entries[model_name] = ReaderUndubEntry(
                reader_jp,
                model_jp,
                model_eu.name,
                model_eu.number,
                new_size=model_jp.size,
                japanese=True,
            )

        return untouched_jp_entries

    def replace_night_titles_jp():
        msn_titles_jp: dict[str, TOCEntry] = {}
        for toc_jp in entries_jp:
            if re.match(r"MSN0[1234]TTL\.PK2", toc_jp.name):
                name, _, ext = toc_jp.name.partition(".")
                for lang in ("E", "F", "G", "S", "I"):
                    eu_name = f"{name}_{lang}.{ext}"
                    msn_titles_jp[eu_name] = toc_jp
        if len(msn_titles_jp) != 5 * 4:
            raise RuntimeError("cannot find night titles images in japanese iso")

        msn_titles_eu = {
            toc_eu.name: toc_eu for toc_eu in entries_eu if re.match(r"MSN0[1234]TTL_[EFGSI]\.PK2", toc_eu.name)
        }
        if len(msn_titles_eu) != 5 * 4:
            raise RuntimeError("cannot find night titles images in european iso")

        NUM_TIM2_IN_TITLE = 11

        new_msn_titles: dict[str, ExternalFileEntry] = {}

        for msn_title_name in msn_titles_eu:
            msn_title_jp = msn_titles_jp[msn_title_name]
            msn_title_eu = msn_titles_eu[msn_title_name]
            with reader_jp.open(msn_title_jp.name) as fh_jp, reader_eu.open(msn_title_eu.name) as fh_eu:
                archive_jp = PK2Archive(fh_jp)
                archive_eu = PK2Archive(fh_eu, copy=True)
                for i in range(NUM_TIM2_IN_TITLE):
                    archive_eu[i] = archive_jp[i]

                new_msn_titles[msn_title_eu.name] = ExternalFileEntry(
                    file=archive_eu.data,
                    name=msn_title_eu.name,
                    number=msn_title_eu.number,
                )

        return new_msn_titles

    scene_audio_entries = filter_jp_entries_common(r"SCENE.*\.STR", pad=True)
    sfx_audio_entries = filter_jp_entries_common(r"^((?!SCENE).).*\.STR", pad=True) if replace_sfx else {}
    bd_audio_entries = filter_jp_entries_common(r"^.*\.BD", pad=True) if replace_sfx else {}
    ingame_text_en = filter_ingame_text_en()
    title_entries = repack_title() if replace_title_jp else {}

    jp_model_entries: dict[str, AbstractUndubEntry] = {}
    if replace_models:
        jp_model_entries.update(repack_pl_mtop())
        jp_model_entries.update(replace_models_untouched())
        jp_model_entries.update(replace_night_titles_jp())

    undub_entries: list[AbstractUndubEntry] = []

    for toc in entries_eu:
        if toc.name == ingame_text_en.name:
            undub_entries.append(patch_english_subtitles(reader_eu, toc))

        elif toc.name in scene_audio_entries:
            undub_entries.append(scene_audio_entries[toc.name])

        elif toc.name in bd_audio_entries:
            undub_entries.append(bd_audio_entries[toc.name])

        elif replace_sfx and toc.name in sfx_audio_entries:
            undub_entries.append(sfx_audio_entries[toc.name])

        elif toc.name in title_entries:
            undub_entries.append(title_entries[toc.name])

        elif toc.name in jp_model_entries:
            undub_entries.append(jp_model_entries[toc.name])

        else:
            undub_entries.append(ReaderUndubEntry(reader_eu, toc, toc.name, toc.number))

    sizes = [
        entry.new_size
        for entry in sorted(undub_entries, key=lambda e: e.number)  # pyright: ignore[reportGeneralTypeIssues]
    ]

    max_img_bd_size = reader_eu.adapter.get_img_bd_size()

    if max_img_bd_size is None:
        raise RuntimeError("cannot get max img_bd size")

    found = False
    offsets: list[int] = []
    img_bd_size = -1
    align_values = [16, 8, 4, 2, 1, 0]
    align = align_values[0]
    while not found and align > 0:
        align = align_values.pop(0)
        offsets, img_bd_size = recalculate_img_bin_offsets(sizes, align=align)
        found = img_bd_size <= max_img_bd_size

    if not found:
        raise RuntimeError("cannot repack img_bd")

    if callback:
        callback(len(undub_entries) + 1)

    iso = pycdlib.PyCdlib()
    iso.open(out_iso_path, mode="rb")

    # ############ write IMG_HD.BIN ############
    img_hd_offset = iso.get_record(iso_path="/IMG_HD.BIN;1").fp_offset

    with open(out_iso_path, "rb+") as iso_fh:
        img_hd_bin = struct.pack(f"<{len(offsets) * 2}I", *list(itertools.chain(*zip(offsets, sizes))))

        iso_fh.seek(img_hd_offset, os.SEEK_SET)
        iso_fh.write(img_hd_bin)

    if callback:
        callback()

    # ############ write IMG_BD.BIN ############
    img_bd_offset = iso.get_record(iso_path="/IMG_BD.BIN;1").fp_offset

    with open(out_iso_path, "rb+") as iso_fh:
        for n, (entry, offset) in enumerate(zip(undub_entries, offsets)):
            # compute offset for current entry and padding value to zero out up to the next entry
            current_offset = offset * 2048
            next_offset = offsets[n + 1] * 2048 if len(offsets) > n + 1 else img_bd_size
            zero_padding = next_offset - (current_offset + cast(int, entry.data_size))

            # write current entry
            iso_fh.seek(img_bd_offset + current_offset, os.SEEK_SET)
            data = 1
            entry_file_offset = 0
            while data:
                entry.seek(entry_file_offset)
                data = entry.read(size=16 * 1024)
                entry_file_offset += len(data)  # pyright: ignore[reportGeneralTypeIssues]
                iso_fh.write(data)  # pyright: ignore[reportGeneralTypeIssues]

            # fill with zeros up to the next entry
            while zero_padding > 0:
                padding_to_write = min(16 * 1024, zero_padding)
                iso_fh.write(bytearray(padding_to_write))
                zero_padding -= padding_to_write

            if callback:
                callback()



from typing import BinaryIO, cast
from contextlib import contextmanager
from pycdlib.pycdlibexception import PyCdlibInvalidISO

from ...dvd import CDDVD
from .abstract import AbstractAdapter


class ISOAdapter(AbstractAdapter):
    def __init__(self, load_path: str):
        super().__init__(load_path)

        try:
            self.iso = CDDVD(load_path)
        except (IsADirectoryError, PermissionError, FileNotFoundError, PyCdlibInvalidISO):
            raise RuntimeError("cannot open iso")  # pylint: disable=raise-missing-from

    def test_file(self, file_path: str) -> bool:
        return bool(self.iso.find_file(file_path))

    def read_file(self, file_name: str, size=-1, offset=0) -> "bytes | None":
        return self.iso.read_file(file_name, size, offset)

    def get_img_bd_size(self) -> "int | None":
        return self.iso.get_file_size(self.img_bd_path)

    def find_file(self, file_name: str) -> "str | None":
        return self.iso.find_file(file_name)

    @contextmanager
    def open(self, file_name: str) -> BinaryIO:  # pyright: ignore[reportGeneralTypeIssues]
        with self.iso.open(file_name) as file_h:
            yield cast(BinaryIO, file_h)



from .utils import make_table


_font_01 = """
 ABCDEFGHIJKLMNOPQRST
UVWXYZabcdefghijklmno
pqrstuvwxyz0123456789
{0}{1}{2}{3}{4}{5}{6}{7}{8}{9}{!?}ÀÂÇÈÉÊÎÔàè
éêîôùûÄẞËÏÖÜäëïü¡¿Á{É1}Í
ÑÓÚá{é1}íñóú{À}{È}{É2}ÌÒÙ{à}{è}{é}ìò{ù}
{Ë}{Ï}Œ{Ch}{LI}{rr}{Rune1}{Rune2}~{Rune3}{Rune4}{Rune5}\"'()-?/⸴⹁
;:,.!「」✓✗{pts}âç°#◯‷=ö#{○}{✕}
{△}{□}壱弐参肆伍陸漆捌玖œ#########
#####################
"""


_font_02 = """
 ABCDEFGHIJKLMNOPQRST
UVWXYZabcdefghijklmno
pqrstuvwxyz0123456789
0123456789{!?}ÀÂÇÈÉÊÎÔàè
éêîôùûÄẞËÏÖÜäëïü¡¿ÁÉÍ
ÑÓÚáéíñóúÀÈÉÌÒÙàèéìòù
ËÏŒ{Ch}{LI}{rr}{Rune1}{Rune2}~{Rune3}{Rune4}{Rune5}\"'()-?/⸴⹁
;:,.!「」✓✗{pts}âç°#◯‷=ö#{○}{✕}
{△}{□}#########œ#########
#####################
"""


_font_03 = """
 ABCDEFGHIJKLMNOPQRST
UVWXYZabcdefghijklmno
pqrstuvwxyz0123456789
0123456789{!?}ÀÂÇÈÉÊÎÔàè
éêîôùûÄẞËÏÖÜäëïü¡¿ÁÉÍ
ÑÓÚáéíñóúÀÈÉÌÒÙàèéìòù
ËÏŒ{Ch}{LI}{rr}###############
#####################
#####################
#####################
"""

from .japanese import _font_04  # noqa: E402 # pylint: disable=wrong-import-position


from .japanese import _font_05  # noqa: E402 # pylint: disable=wrong-import-position


table_eu = {
    "default": make_table(_font_01),
    0xF0: make_table(_font_02),
    0xF1: make_table(_font_03),
    0xF2: make_table(_font_04),
    0xF3: make_table(_font_05),
}



import io
import os
import re
import struct

from abc import ABC, abstractmethod
from typing import BinaryIO

from .tables.subtitles import decode_english_subtitles
from .tables import table_eu, table_jp, LASTCH, COLOR, NEWLINE


class Serializable(ABC):
    @abstractmethod
    def encode(self, offset=0):
        ...


class InGameMessageTable(Serializable):
    def __init__(
        self,
        number: int,
        offset: int,
        size: int,
        parent: "InGameMessageTable | None" = None,
        tables: "list[InGameMessageTable] | None" = None,
        messages: "list[InGameMessage] | None" = None,
    ):
        self.number = number
        self.offset = offset
        self.size = size
        self._parent = parent
        self.tables = tables if tables is not None else []
        self.messages: list["InGameMessage"] = messages if messages is not None else []

    def add_table(self, table: "InGameMessageTable"):
        if self.messages:
            raise RuntimeError("cannot add table while messages exist")

        self.tables.append(table)

    def add_message(self, message: "InGameMessage"):
        if self.tables:
            raise RuntimeError("cannot add message while tables exists")

        self.messages.append(message)

    def __str__(self):
        if self.tables:
            return f'{len(self.tables)} Table{"s" if len(self.tables) != 1 else ""}'
        else:
            return f'{len(self.messages)} Message{"s" if len(self.messages) != 1 else ""}'

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        return id(self) == id(other)

    def __getitem__(self, item):
        if self.tables:
            return self.tables[item]
        else:
            return self.messages[item]

    def has_overlap(self, other):
        r1 = range(self.offset, self.offset + self.size)
        r2 = range(other.offset, other.offset + other.size)
        return len(list(set(r1) & set(r2))) != 0

    def encode(self, offset=0):
        table = self.tables if self.tables else self.messages

        encoded = io.BytesIO()
        offsets = []

        encoded.write(bytearray(len(table * 4)))
        offset += len(table) * 4

        for entry in table:
            offsets.append(offset)
            enc = entry.encode(offset)
            encoded.write(enc)
            offset += len(enc)

        encoded.seek(0, os.SEEK_SET)

        table_bin = struct.pack(f"<{len(offsets)}I", *offsets)
        encoded.write(table_bin)

        encoded.seek(0, os.SEEK_SET)

        return encoded.read()


class InGameMessage(Serializable):
    def __init__(
        self, number: int, offset: int, parent: "InGameMessageTable | None" = None, japanese: "bool | None" = False
    ):
        self.number: int = number
        self.offset: int = offset
        self._parent: "InGameMessageTable | None" = parent
        self.size: int = 0
        self.message: str = ""
        self.suffix: bytes = b""
        self.data: bytes = b""
        self.data_hex_str: str = ""
        self.japanese = japanese

    @classmethod
    def from_message(
        cls, number: int, parent: InGameMessageTable, message: str, offset=-1, suffix=b"\xff", japanese=False
    ):
        msg = cls(number, offset, parent, japanese)
        msg.message = message
        msg.suffix = suffix
        msg.encode()
        return msg

    @classmethod
    def from_data(cls, number, parent, data: bytes, offset=-1, japanese=False):
        msg = cls(number, offset, parent, japanese)
        msg._parse_data(data)
        return msg

    @staticmethod
    def _maybe_japanese(data):
        has_jp_symbols_iter = (True for x in data if x in table_jp)
        return next(has_jp_symbols_iter, False)

    @staticmethod
    def _separate_suffix(data: bytes) -> tuple[bytes, bytes]:
        match = re.match(b"^(.*?)(\xFA?\xFF*)$", data, re.DOTALL)
        assert match is not None
        data, suffix = match.groups()
        return data, suffix

    def parse_text(self, file_h: BinaryIO, size: int):
        file_h.seek(self.offset)
        data = file_h.read(size)
        return self._parse_data(data)

    def _parse_data(self, data: bytes):
        self.size = len(data)
        self.data_hex_str = " ".join(f"{x:02X}" for x in data)
        self.data, self.suffix = self._separate_suffix(data)

        text_bin = self.data

        if self.japanese is None:
            self.japanese = self._maybe_japanese(text_bin)

        font_table = table_eu if not self.japanese else table_jp

        idx = 0
        while idx < len(text_bin):
            x = text_bin[idx]

            if x == COLOR:
                idx_left = len(text_bin) - 1 - idx
                if idx_left >= 3:
                    # pylint: disable=consider-using-f-string
                    self.message += "{Color#%02X%02X%02X}" % (text_bin[idx + 1], text_bin[idx + 2], text_bin[idx + 3])
                    idx += 3
                else:
                    self.message += "{Color}"

            elif x == NEWLINE:
                self.message += "\n"

            elif self.japanese and x in font_table:
                table = font_table[x]
                idx += 1
                self.message += table[text_bin[idx]]

            elif x <= LASTCH:
                self.message += font_table["default"][x]

            else:
                self.message += f"{{0x{x:02X}}}"

            idx += 1

    @staticmethod
    def _encode_color(color_str):
        return re.sub(
            r"{Color#([A-F0-9]{2})([A-F0-9]{2})([A-F0-9]{2})}",
            f"{{0x{COLOR:02X}}}{{0x\\1}}{{0x\\2}}{{0x\\3}}",
            color_str,
        )

    def encode(self, offset=0):
        message = (
            self._encode_color(self.message)
            .replace("{Color}", f"{{0x{COLOR:02X}}}")
            .replace("\n", f"{{0x{NEWLINE:02X}}}")
        )

        encoded = io.BytesIO()

        char_table = table_jp if self.japanese is True else table_eu

        characters = re.findall(r"({0x[A-F0-9]{2}}|{.*?}|.)", message)
        for ch in characters:
            enc = None
            if len(ch) != 6:
                for name, table in char_table.items():
                    if ch in table:
                        enc = table.index(ch).to_bytes(1, "little")
                    if enc is not None and name != "default" and isinstance(name, int):
                        enc = name.to_bytes(1, "little") + enc
                    if enc is not None:
                        break
            else:
                enc = bytes.fromhex(ch[3:5])

            assert enc is not None

            encoded.write(enc)

        encoded.write(self.suffix)

        encoded.seek(0, os.SEEK_SET)
        encoded = encoded.read()
        encoded_hex_str = " ".join(f"{x:02X}" for x in encoded)

        self.data, self.suffix = self._separate_suffix(encoded)
        self.size = len(encoded)
        self.data_hex_str = encoded_hex_str

        return encoded

    def __str__(self):
        suffix_hex = "".join(f"{x:02X}" for x in self.suffix)
        return f'"{self.message}" ({suffix_hex})'

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        if isinstance(other, InGameMessage):
            return self.offset == other.offset and self.size == other.size
        else:
            return id(self) == id(other)

    def has_overlap(self, other):
        r1 = range(self.offset, self.offset + self.size)
        r2 = range(other.offset, other.offset + other.size)
        return len(list(set(r1) & set(r2))) != 0


class InGameMessageParser:
    def __init__(self, file: "str | bytes | BinaryIO", table_names: "list[str] | None" = None, japanese=False):
        if isinstance(file, str):
            with open(file, "rb") as fh:
                self.file_h = io.BytesIO(fh.read())
        elif isinstance(file, bytes):
            self.file_h = io.BytesIO(file)
        elif isinstance(file, io.IOBase):
            pos = file.tell()
            file.seek(0, os.SEEK_SET)
            self.file_h = io.BytesIO(file.read())
            file.seek(pos, os.SEEK_SET)
        else:
            raise ValueError("input must be filename or open file handle")

        self.file_size = self.file_h.getbuffer().nbytes
        self.japanese = japanese

        self._boundaries: list[int] = []
        self._byte_address_read = []

        self.msg_tables = self._parse_obj(japanese=japanese)

        self.table_names = None

        if isinstance(table_names, list) and len(table_names) == len(self.msg_tables.tables):
            self.table_names = table_names

    def __getitem__(self, item):
        assert self.msg_tables is not None

        if isinstance(item, int):
            return self.msg_tables.tables[item]

        assert self.table_names is not None
        table_idx = self.table_names.index(item)
        return self.msg_tables.tables[table_idx]

    @staticmethod
    def _read_dword(f):
        return struct.unpack("<I", f.read(4))[0], f.tell()

    def next_boundary(self, offset):
        return next(b for b in self._boundaries if b > offset)

    def _parse_obj_rec(self, msg_table: InGameMessageTable):
        file_has_table = len(msg_table.tables) > 0

        if file_has_table:
            for sub_table in msg_table.tables:
                self._parse_obj_rec(sub_table)

            return

        for message in msg_table.messages:
            size = self.next_boundary(message.offset) - message.offset
            addresses = [x for x in range(message.offset, message.offset + message.size)]
            self._byte_address_read.extend(addresses)
            message.parse_text(self.file_h, size)
            message.encode()

    def _find_table_size(self, offset):
        max_offset = self.file_size
        position = -1
        maybe_table = []
        address_read = []
        while max_offset > position:
            address_read.extend([x for x in range(self.file_h.tell(), self.file_h.tell() + 4)])
            if self.file_h.tell() + 4 >= self.file_size:
                return False, 0
            address, position = self._read_dword(self.file_h)
            max_offset = min(address, max_offset)
            if address > self.file_size or max_offset < offset:
                return False, 0
            maybe_table.append(address)
        self._byte_address_read.extend(address_read)
        return maybe_table, max_offset - offset

    def _parse_obj_tables(
        self, number, offset=0, msg_table: "InGameMessageTable | None" = None, japanese: "bool | None" = False
    ):
        self.file_h.seek(offset, os.SEEK_SET)

        table, tbl_size = self._find_table_size(offset)

        if table is False:
            return

        cur_msg_table = InGameMessageTable(number, offset, tbl_size, msg_table)
        self._boundaries.append(offset)

        if msg_table is not None:
            msg_table.add_table(cur_msg_table)

        for n, offset_start in enumerate(table):
            has_subtable = self._parse_obj_tables(n, offset_start, cur_msg_table, japanese) is not None
            if not has_subtable:
                cur_msg_table.add_message(InGameMessage(n, offset_start, cur_msg_table, japanese))
                self._boundaries.append(offset_start)

        return cur_msg_table

    def _parse_obj(self, japanese: "bool | None" = None):
        msg_tables = self._parse_obj_tables(number=0, japanese=japanese)
        assert msg_tables is not None

        self._boundaries.append(self.file_size)
        self._boundaries = list(set(self._boundaries))
        self._boundaries.sort()

        for table in msg_tables.tables:
            self._parse_obj_rec(table)

        return msg_tables

    def encode(self):
        assert self.msg_tables is not None
        return self.msg_tables.encode()


def inject_english_subtitles(file_h: BinaryIO):
    ig_msg_parser = InGameMessageParser(file_h, japanese=False)

    parent = ig_msg_parser[53]
    parent.messages = []

    subtitles_en = decode_english_subtitles(file_h)

    for n, sub in subtitles_en.items():
        msg = InGameMessage.from_message(n, parent, sub)
        parent.messages.append(msg)

    return ig_msg_parser.encode()



import os
import io
import errno

from typing import BinaryIO


class SubFile(io.BufferedIOBase, BinaryIO):  # pylint: disable=abstract-method
    def __init__(self, file_h: BinaryIO, offset: int, size: int):
        self.file_h: BinaryIO = file_h
        self.offset: int = offset
        self.size: int = size
        self.position: int = 0

    def tell(self):
        return self.position

    def seek(self, position: int, whence=os.SEEK_SET):
        if whence == os.SEEK_SET:
            next_position = min(self.size, position)
        elif whence == os.SEEK_CUR:
            next_position = min(self.size, self.position + position)
        elif whence == os.SEEK_END:
            next_position = min(self.size, self.size + position)
        else:
            raise OSError(errno.ENXIO, os.strerror(errno.ENXIO))

        if next_position < 0:
            raise OSError(errno.EINVAL, os.strerror(errno.EINVAL))

        self.position = next_position

        return self.position

    def read(self, size=-1, offset=0):
        if offset > 0:
            self.seek(self.position + offset)

        if (isinstance(size, int) and size == -1) or size is None:
            size = self.size - self.position
        elif isinstance(size, int) and size >= 0:
            size = min(size, self.size - self.position)
        else:
            raise TypeError(f"argument should be integer or None, not '{type(size)}'")

        self.file_h.seek(self.offset + self.position)
        self.position += size

        return self.file_h.read(size)

    def write(self, data):
        self.file_h.seek(self.offset + self.position)
        bytes_to_write = min(self.size - self.position, len(data))
        data_to_write = data[:bytes_to_write]
        self.file_h.write(data_to_write)
        self.position += bytes_to_write



import os
import io
import math
import struct

from typing import BinaryIO, Union
from dataclasses import dataclass

from ..utils.file import SubFile


data_types = {b"TIM2": "TM2", b"\x50\x10\x00\x00": "SGD"}


def parse_pk2_archive(pk2_file_h: BinaryIO):
    raise RuntimeError()


@dataclass
class PK2ArchiveEntry:
    offset: int
    size: int
    type: str


class PK2Archive:
    def __init__(self, pk2_data: BinaryIO, copy: bool = False):
        if copy:
            pk2_data = io.BytesIO(pk2_data.read())

        self.data = pk2_data
        self.entries: list[PK2ArchiveEntry] = []

        sz = pk2_data.seek(0, os.SEEK_END)
        pk2_data.seek(0, os.SEEK_SET)

        num_files: int = struct.unpack("<4I", pk2_data.read(16))[0]
        num_byte_addrs = int(math.ceil(num_files / 4) * 4) * 4

        file_addrs = pk2_data.read(num_byte_addrs)[: num_files * 4]
        offsets: list[int] = list(struct.unpack(f"<{num_files}I", file_addrs))

        sequential = len(offsets) > 1 and offsets[1] == 0

        if not sequential:
            sizes = [offsets[i + 1] - offsets[i] for i in range(len(offsets) - 1)]
            sizes.append(sz - offsets[-1])
            for addr, size in zip(offsets, sizes):
                pk2_data.seek(addr, os.SEEK_SET)
                data_type = data_types.get(pk2_data.read(4), "BIN")
                self.entries.append(PK2ArchiveEntry(addr, size, data_type))
        else:
            next_addr: int = 0x0
            while next_addr + 0x20 < sz:
                next_addr += 0x10
                pk2_data.seek(next_addr)
                file_size: int = struct.unpack("<4I", pk2_data.read(16))[0]
                file_addr = pk2_data.tell()
                next_addr += file_size
                data_type = data_types.get(pk2_data.read(4), "BIN")
                self.entries.append(PK2ArchiveEntry(file_addr, file_size, data_type))

            assert len(self.entries) == num_files

    def __len__(self):
        return len(self.entries)

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    def __getitem__(self, item: int):
        assert item < len(self.entries)
        entry = self.entries[item]
        return SubFile(self.data, entry.offset, entry.size)

    def __setitem__(self, item: int, data: Union[bytes, bytearray, SubFile]):
        assert item < len(self.entries)
        entry = self.entries[item]
        if isinstance(data, SubFile):
            data.seek(0, os.SEEK_SET)
            data_bin = data.read()
        else:
            data_bin = data
        assert len(data_bin) == entry.size
        self.data.seek(entry.offset, os.SEEK_SET)
        self.data.write(data_bin)


def main():
    for lang in ("E", "F", "G", "I", "S"):
        pk2_file = f"/home/baecchi/Documents/PyCharm/ProjectZeroUndub/MODELS/PL_MTOP_{lang}.PK2"

        with open(pk2_file, "rb") as file_h:
            pk2_archive = PK2Archive(file_h)
            for i, entry in enumerate(pk2_archive):
                data = entry.read()
                out_dir_path = os.path.splitext(pk2_file)[0]
                os.makedirs(out_dir_path, exist_ok=True)
                with open(os.path.join(out_dir_path, f"file{i:04d}.tm2"), "wb") as tm2_out_h:
                    tm2_out_h.write(data)


if __name__ == "__main__":
    main()



import os
import sys
import tqdm

from typing import Type

from .error import AbstractProgressError


# noinspection PyBroadException
def remove_ctrl_c_echo():
    try:
        if "linux" in sys.platform and sys.stdout.isatty():
            os.system("stty -echoctl")
    except Exception:
        pass


class Progress:
    def __init__(self, desc: str, error: Type[AbstractProgressError], tpad=20, ncols=50):
        self.error = error
        self.pbar = tqdm.tqdm(
            bar_format=f"{{desc}}: {{percentage:3.0f}}% ┤{{bar:{ncols}}}├ {{n_fmt}}/{{total_fmt}}{{bar:-{ncols}b}}",
            desc=f"  \x1b[93m{desc:{tpad}s}\x1b[0m",
            smoothing=1,
            colour="blue",
            file=sys.stdout,
        )

    def __enter__(self):
        remove_ctrl_c_echo()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if exc_type is not None or self.pbar.n < self.pbar.total:
            self.pbar.colour = "red"
            self.pbar.refresh()
            self.pbar.close()

            self.error.print()
            sys.exit(1)

        if self.pbar.n > self.pbar.total:
            self.pbar.n = self.pbar.total

        self.pbar.colour = "green"
        self.pbar.refresh()
        self.pbar.close()

    def __call__(self, total=None):
        if total:
            self.pbar.total = total
            self.pbar.n = 0
        else:
            self.pbar.update(1)

        self.pbar.refresh()



import sys
from abc import ABC
from dataclasses import dataclass


@dataclass
class AbstractProgressError(ABC):
    error_msg: str
    explanation: str
    suggestion: str

    def __post_init__(self):
        raise RuntimeError("Cannot instantiate ProgressError")

    @classmethod
    def print(cls, file=sys.stdout):
        file.flush()
        file.write("\n")
        file.write("  " + cls.error_msg.replace("Error", "\x1b[1;31mError\x1b[0m") + ".\n")
        file.write("  " + f"\x1b[4m{cls.explanation}\x1b[0m" + ".\n")
        file.write("  " + cls.suggestion + ".\n")
        file.write("\n")
        file.flush()



import os
import io
import scipy
import struct
import numpy as np

from typing import Union, BinaryIO
from functools import lru_cache


class InvalidTim2FormatException(Exception):
    pass


class Tim2ImageData:
    bpp_map = {1: 16, 2: 24, 3: 32, 4: 4, 5: 8}
    bpp_map_inv = {v: k for k, v in bpp_map.items()}
    header_size = 48

    def __init__(
        self,
        total_length: int,
        palette_length: int,
        data_length: int,
        header_length: int,
        color_entries: int,
        image_format: int,
        mipmap_count: int,
        clut_format: int,
        bbp: int,
        width: int,
        height: int,
        gs_tex0: int,
        gs_tex1: int,
        gs_regs: int,
        gs_tex_clut: int,
    ):
        self.total_length: int = total_length  # 4 - Total Image Length
        self.palette_length: int = palette_length  # 4 - Palette Length
        self.data_length: int = data_length  # 4 - Image Data Length
        self.header_length: int = header_length  # 2 - Header Length
        self.color_entries: int = color_entries  # 2 - Color Entries
        self.image_format: int = image_format  # 1 - Image Format (0=8bpp w/ palette?)
        self.mipmap_count: int = mipmap_count  # 1 - Mipmap Count
        self.clut_format: int = clut_format  # 1 - CLUT Format
        self.bpp: int = bbp  # 1 - Bits Per Pixel (1=16bbp, 2=24bpp, 3=32bbp, 4=4bbp, 5=8bpp)
        self.width: int = width  # 2 - Image Width
        self.height: int = height  # 2 - Image Height
        self.gs_tex0: int = gs_tex0  # 8 - GsTEX0
        self.gs_tex1: int = gs_tex1  # 8 - GsTEX1
        self.gs_regs: int = gs_regs  # 4 - GsRegs
        self.gs_tex_clut: int = gs_tex_clut  # 4 - GsTexClut

        self.linear_palette: bool = False

        self.data: Union[bytearray, None] = None  # X - User Data (optional) (length = HeaderLength - 48)
        self.palette: Union[bytearray, None] = None

    def post_init(self, file_handle: BinaryIO):
        self.bpp = Tim2ImageData.bpp_map.get(self.bpp, self.bpp)

        data_len = self.total_length - Tim2ImageData.header_size

        data = file_handle.read(data_len)

        if len(data) != data_len:
            raise RuntimeError("wrong tim2 texture format")

        self.data = bytearray(data[: self.data_length])
        self.palette = bytearray(data[self.data_length :])

        if len(self.palette) != self.palette_length:
            raise InvalidTim2FormatException("unexpected palette size")

        self.linear_palette = self.clut_format & 0x80 != 0
        self.clut_format &= 0x7F


class Tim2Image:
    def __init__(self, file_handle):
        self.signature: bytes  # 4 - Signature (TIM2)
        self.version: int  # 2 - Version
        self.num_images: int  # 2 - Number of Images

        self.image: Union[Tim2ImageData, None] = None

        self.post_init(file_handle)

    def post_init(self, file_handle):
        sz = file_handle.seek(0, os.SEEK_END)
        file_handle.seek(0, os.SEEK_SET)

        if sz < 56:
            raise InvalidTim2FormatException("header too small")

        header = file_handle.read(16)
        if header[:4] != b"TIM2":
            raise InvalidTim2FormatException("invalid signature")

        self.signature = header
        self.version, self.num_images = struct.unpack("<HH", header[4:8])

        if self.num_images != 1:
            raise InvalidTim2FormatException("only one image per tim2 supported")

        img_meta = file_handle.read(48)

        if len(img_meta) != 48:
            raise InvalidTim2FormatException("not enough data to parse")

        tim2_image_data = Tim2ImageData(*struct.unpack("<IIIHHBBBBHHQQII", img_meta))
        tim2_image_data.post_init(file_handle)

        self.image = tim2_image_data

    def to_bytes(self):
        assert self.image and self.image.data and self.image.palette
        out_handle = io.BytesIO()
        out_handle.write(b"TIM2")
        out_handle.write(struct.pack("<H", self.version))
        out_handle.write(struct.pack("<H", self.num_images))
        out_handle.write(b"\x00" * 8)
        out_handle.write(struct.pack("<I", self.image.total_length))
        out_handle.write(struct.pack("<I", self.image.palette_length))
        out_handle.write(struct.pack("<I", self.image.data_length))
        out_handle.write(struct.pack("<H", self.image.header_length))
        out_handle.write(struct.pack("<H", self.image.color_entries))
        out_handle.write(struct.pack("<B", self.image.image_format))
        out_handle.write(struct.pack("<B", self.image.mipmap_count))
        out_handle.write(struct.pack("<B", self.image.clut_format | (self.image.linear_palette * 0x80)))
        out_handle.write(struct.pack("<B", self.image.bpp_map_inv[self.image.bpp]))
        out_handle.write(struct.pack("<H", self.image.width))
        out_handle.write(struct.pack("<H", self.image.height))
        out_handle.write(struct.pack("<Q", self.image.gs_tex0))
        out_handle.write(struct.pack("<Q", self.image.gs_tex1))
        out_handle.write(struct.pack("<I", self.image.gs_regs))
        out_handle.write(struct.pack("<I", self.image.gs_tex_clut))
        out_handle.write(self.image.data)
        out_handle.write(self.image.palette)
        return out_handle.getvalue()


first_row = 60
idx_start = (
    192,
    192,
    189,
    187,
    185,
    183,
    182,
    181,
    180,
    179,
    178,
    177,
    176,
    176,
    175,
    175,
    174,
    174,
    174,
    172,
    172,
    172,
    172,
    172,
    172,
    172,
    172,
    172,
    172,
    172,
    172,
    172,
    172,
    174,
    174,
    174,
    175,
    175,
    176,
    176,
    177,
    178,
    179,
    180,
    181,
    182,
    183,
    185,
    187,
    189,
    192,
    192,
)
idx_num = (
    15,
    15,
    21,
    25,
    29,
    33,
    35,
    37,
    39,
    41,
    43,
    45,
    47,
    47,
    49,
    49,
    51,
    51,
    51,
    55,
    55,
    55,
    55,
    55,
    55,
    55,
    55,
    55,
    55,
    55,
    55,
    55,
    55,
    51,
    51,
    51,
    49,
    49,
    47,
    47,
    45,
    43,
    41,
    39,
    37,
    35,
    33,
    29,
    25,
    21,
    15,
    15,
)


def patch_pl_mtop(tim2_eu_io: BinaryIO, tim2_jp_io: BinaryIO):
    tim2img_eu = Tim2Image(tim2_eu_io)
    assert tim2img_eu.image and tim2img_eu.image.data and tim2img_eu.image.palette

    tim2img_jp = Tim2Image(tim2_jp_io)
    assert tim2img_jp.image and tim2img_jp.image.data and tim2img_jp.image.palette

    indices = []
    for i, (s, n) in enumerate(zip(idx_start, idx_num)):
        r = (first_row + i) * tim2img_eu.image.width
        indices.extend(list(range(r + s, r + s + n + 1)))

    p_eu = list(struct.iter_unpack("<4B", tim2img_eu.image.palette))
    p_jp = list(struct.iter_unpack("<4B", tim2img_jp.image.palette))

    refilter_map = None
    if not tim2img_eu.image.linear_palette:
        p_eu, refilter_map = defilter_palette(p_eu)
    if not tim2img_jp.image.linear_palette:
        p_jp, _ = defilter_palette(p_jp)

    def np_take(pal: list, arr: bytearray):
        return np.take(pal, np.frombuffer(arr, dtype=np.uint8), axis=0).astype(np.uint8)

    img_eu = np_take(p_eu, tim2img_eu.image.data)
    img_jp = np_take(p_jp, tim2img_jp.image.data)

    def unique_colors(img: np.ndarray):
        return np.unique(img, axis=0)

    img_eu[indices] = img_eu[-1]
    colors_eu = unique_colors(img_eu)
    p_eu = colors_eu.tolist()

    color_errors: dict[tuple, list] = {}
    for idx in indices:
        jp_col = img_jp[idx]
        dists = np.power(p_eu - jp_col, 2).sum(axis=1)
        min_c_idx = dists.argmin()
        jp_col_tp = tuple(jp_col.tolist())
        if jp_col_tp not in color_errors:
            color_errors[jp_col_tp] = [0, []]
        color_errors[jp_col_tp][0] = dists[min_c_idx]
        color_errors[jp_col_tp][1].append(idx)
        img_eu[idx] = p_eu[min_c_idx]

    assert len(np.unique(img_eu, axis=0)) <= 256
    assert len(np.unique(img_jp, axis=0)) <= 256

    new_colors = unique_colors(img_eu)
    free_colors = 256 - len(new_colors)

    new_color_candidates = sorted(color_errors.items(), key=lambda x: x[1][0], reverse=True)[:free_colors]

    for color, (_, where_indices) in new_color_candidates:
        for idx in where_indices:
            img_eu[idx] = color

    new_palette_eu = unique_colors(img_eu)

    assert len(new_palette_eu) == 256

    @lru_cache(maxsize=None)
    def to_gray(color):
        r, g, b = color[:3]
        return (0.299 * r) + (0.587 * g) + (0.114 * b)

    new_palette_eu = sorted(new_palette_eu.tolist(), key=lambda x: to_gray(tuple(x)))

    img_eu_palettized = np.where(scipy.spatial.distance.cdist(img_eu, new_palette_eu) == 0)[1].astype(np.uint8)

    # refilter palette
    if not tim2img_eu.image.linear_palette:
        new_palette_eu = refilter_palette(new_palette_eu, refilter_map)

    or_palette_len = len(tim2img_eu.image.palette)
    or_data_len = len(tim2img_eu.image.data)

    tim2img_eu.image.palette = bytearray(np.array(new_palette_eu, dtype=np.uint8).tobytes())
    tim2img_eu.image.data = bytearray(img_eu_palettized.tobytes())

    assert len(tim2img_eu.image.palette) == or_palette_len
    assert len(tim2img_eu.image.data) == or_data_len

    return tim2img_eu.to_bytes()


def refilter_palette(palette, refilter_map=None):
    if refilter_map is None:
        _, refilter_map = defilter_palette(palette)

    length = len(palette)

    new_colors = [0] * length

    for i in range(length):
        new_colors[refilter_map[i]] = palette[i]

    return new_colors


def defilter_palette(palette):
    length = len(palette)

    parts = length // 32
    stripes = 2
    colors = 8
    blocks = 2

    new_colors = [0] * length

    refilter_map: list[int] = []

    i = 0
    for part in range(parts):
        for block in range(blocks):
            for stripe in range(stripes):
                for color in range(colors):
                    j = (part * colors * stripes * blocks) + (block * colors) + (stripe * stripes * colors) + color
                    new_colors[i] = palette[j]
                    refilter_map.append(j)
                    i += 1

    return new_colors, refilter_map



import os
import re
import sys
import inspect
import argparse
import threading
import tkinter as tk

from typing import Type
from tkinter import ttk
from tkinter import filedialog as fd
from pycdlib.pycdlibexception import PyCdlibInvalidInput

from zeroundub.cli.utils import copy_file
from zeroundub.zero.iso import (
    check_iso_hashes,
    replace_movies_in_iso_inplace,
    patch_elf_inplace,
    merge_iso_img_bd_contents,
)

from zeroundub.cli.error import AbstractProgressError
from zeroundub.cli.cmdline.undub import (
    ProgressErrorCRCEU,
    ProgressErrorCRCJP,
    ProgressErrorISOCopy,
    ProgressErrorMoviePatch,
    ProgressErrorELFPatch,
    ProgressErrorGamePatch,
)


class Progress:
    # ignore: reportOptionalSubscript
    def __init__(self, app: "App", label_name: ttk.Label, label_perc: ttk.Label, error: Type[AbstractProgressError]):
        self.app = app
        self.label_name = label_name
        self.label_perc = label_perc
        self.error = error
        self._make_red = getattr(self.app, "_make_red")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if exc_type is not None:
            self._make_red(self.label_name)
            self._make_red(self.label_perc)
            self._make_red(self.app.label_info)
            self.app.label_info["text"] = (
                ".\n\n".join((self.error.error_msg, self.error.explanation, self.error.suggestion)) + "."
            )


def format_help_text(text):
    return re.sub(r"^(:?( *)$|( *))", r"", text.strip(), flags=re.MULTILINE).replace("$", " ")


class UndubThread(threading.Thread):
    def __init__(self, app: "App"):
        super().__init__()
        self.app = app
        self.total = 0
        self.current = 0

    def __getattribute__(self, name: str):
        attribute = super().__getattribute__(name)

        if inspect.ismethod(attribute) and name.startswith("do_"):

            def outer(*args, **kwargs):
                def inner():
                    attribute(*args, **kwargs)

                self.app.after(0, inner)

            return outer

        return attribute

    def do_close(self):
        self.app.quit()

    def _update_progress(self, var_progress: tk.DoubleVar, label: ttk.Label, label_perc: ttk.Label, total=None):
        if total:
            self.current = 0
            self.total = total
            label.state(["!disabled"])
            label_perc.state(["!disabled"])
            label_perc["text"] = f"0/{total}"
            var_progress.set(0)
        else:
            self.current += 1
            value = 100 * self.current / self.total
            label_perc["text"] = f"{self.current}/{self.total}"
            var_progress.set(value)

    def do_eu_check_cb(self, total=None):
        self._update_progress(
            self.app.var_progress_eu_check, self.app.label_eu_check, self.app.label_eu_check_perc, total
        )

    def do_jp_check_cb(self, total=None):
        self._update_progress(
            self.app.var_progress_jp_check, self.app.label_jp_check, self.app.label_jp_check_perc, total
        )

    def do_copy_file_cp(self, total=None):
        self._update_progress(self.app.var_progress_copy, self.app.label_copy, self.app.label_copy_perc, total)

    def do_replace_movies_cp(self, total=None):
        self._update_progress(self.app.var_progress_movies, self.app.label_movies, self.app.label_movies_perc, total)

    def do_patch_elf_cp(self, total=None):
        self._update_progress(self.app.var_progress_elf, self.app.label_elf, self.app.label_elf_perc, total)

    def do_merge_iso_cp(self, total=None):
        self._update_progress(
            self.app.var_progress_contents, self.app.label_contents, self.app.label_contents_perc, total
        )

    def do_done(self):
        self.app.label_info["text"] = format_help_text(
            """
            Undubbing completed successfully.

            Please enjoy the game.
        """
        )

        self.app.button_done.state(["!disabled"])

    def run(self) -> None:
        try:
            with Progress(self.app, self.app.label_eu_check, self.app.label_eu_check_perc, ProgressErrorCRCEU):
                check_iso_hashes(
                    self.app.eu_iso_path,
                    lang="EU",
                    callback=self.do_eu_check_cb,
                )

            with Progress(self.app, self.app.label_jp_check, self.app.label_jp_check_perc, ProgressErrorCRCJP):
                check_iso_hashes(
                    self.app.jp_iso_path,
                    lang="JP",
                    callback=self.do_jp_check_cb,
                )

            with Progress(self.app, self.app.label_copy, self.app.label_copy_perc, ProgressErrorISOCopy):
                copy_file(
                    self.app.eu_iso_path,
                    self.app.undub_iso_path,
                    self.do_copy_file_cp,
                )

            with Progress(self.app, self.app.label_movies, self.app.label_movies_perc, ProgressErrorMoviePatch):
                replace_movies_in_iso_inplace(
                    self.app.jp_iso_path,
                    self.app.undub_iso_path,
                    self.do_replace_movies_cp,
                )

            with Progress(self.app, self.app.label_elf, self.app.label_elf_perc, ProgressErrorELFPatch):
                patch_elf_inplace(
                    self.app.undub_iso_path,
                    fix_kirie_camera_bug=self.app.var_fix_kirie_camera_bug.get(),
                    force_lang=self.app.var_force_language_selection.get(),
                    no_bloom=self.app.var_disable_bloom.get(),
                    dark_filter=self.app.var_remove_dark_filter.get(),
                    ingame_noise=self.app.var_remove_ingame_noise.get(),
                    menu_noise=self.app.var_remove_title_noise.get(),
                    force_16_9_game=self.app.var_16_9.get(),
                    force_16_9_movies=self.app.var_16_9.get(),
                    callback=self.do_patch_elf_cp,
                )

            with Progress(self.app, self.app.label_contents, self.app.label_contents_perc, ProgressErrorGamePatch):
                merge_iso_img_bd_contents(
                    eu_iso_path=self.app.eu_iso_path,
                    jp_iso_path=self.app.jp_iso_path,
                    out_iso_path=self.app.undub_iso_path,
                    replace_title_jp=self.app.var_replace_title_jp.get(),
                    replace_models=self.app.var_replace_models.get(),
                    replace_sfx=True,
                    callback=self.do_merge_iso_cp,
                )

            self.do_done()
        except (RuntimeError, PyCdlibInvalidInput):
            return


class App(ttk.Frame):
    filetypes = (("ISO Files", "*.iso"),)

    # noinspection PyUnusedLocal
    def __init__(self, parent):
        ttk.Frame.__init__(self)

        self._undub_started = False

        self.patch_frame: ttk.LabelFrame

        self.check_fix_kirie_camera_bug: ttk.Checkbutton
        self.check_16_9: ttk.Checkbutton
        self.check_disable_bloom: ttk.Checkbutton
        self.check_remove_dark_filter: ttk.Checkbutton
        self.check_remove_ingame_noise: ttk.Checkbutton
        self.check_remove_title_noise: ttk.Checkbutton
        self.check_force_language_selection: ttk.Checkbutton
        self.check_replace_title_jp: ttk.Checkbutton
        self.check_replace_models: ttk.Checkbutton

        self.widgets_frame: ttk.Frame

        self.button_iso_eu: ttk.Button
        self.button_iso_jp: ttk.Button
        self.button_iso_undub: ttk.Button

        self.progress_frame: ttk.Frame
        self.pb_frame_box: ttk.Frame

        self.button_start_undub: ttk.Button

        self.label_eu_check: ttk.Label
        self.label_jp_check: ttk.Label
        self.label_copy: ttk.Label
        self.label_movies: ttk.Label
        self.label_elf: ttk.Label
        self.label_contents: ttk.Label

        self.label_eu_check_perc: ttk.Label
        self.label_jp_check_perc: ttk.Label
        self.label_copy_perc: ttk.Label
        self.label_movies_perc: ttk.Label
        self.label_elf_perc: ttk.Label
        self.label_contents_perc: ttk.Label

        self.progress_eu_check: ttk.Progressbar
        self.progress_jp_check: ttk.Progressbar
        self.progress_copy: ttk.Progressbar
        self.progress_movies: ttk.Progressbar
        self.progress_elf: ttk.Progressbar
        self.progress_contents: ttk.Progressbar

        self.button_done: ttk.Button

        self.info_frame: ttk.LabelFrame
        self.label_info: ttk.Label

        self.sizegrip: ttk.Sizegrip

        # Create control variables
        self.var_fix_kirie_camera_bug = tk.BooleanVar(value=True)
        self.var_16_9 = tk.BooleanVar()
        self.var_disable_bloom = tk.BooleanVar()
        self.var_remove_dark_filter = tk.BooleanVar()
        self.var_remove_ingame_noise = tk.BooleanVar()
        self.var_remove_title_noise = tk.BooleanVar()
        self.var_force_language_selection = tk.BooleanVar()
        self.var_replace_title_jp = tk.BooleanVar()
        self.var_replace_models = tk.BooleanVar()

        self.var_progress_eu_check = tk.DoubleVar(value=0.0)  # 0.0 - 100.0
        self.var_progress_jp_check = tk.DoubleVar(value=0.0)  # 0.0 - 100.0
        self.var_progress_copy = tk.DoubleVar(value=0.0)  # 0.0 - 100.0
        self.var_progress_movies = tk.DoubleVar(value=0.0)  # 0.0 - 100.0
        self.var_progress_elf = tk.DoubleVar(value=0.0)  # 0.0 - 100.0
        self.var_progress_contents = tk.DoubleVar(value=0.0)  # 0.0 - 100.0

        # variables
        self.eu_iso_path: str
        self.jp_iso_path: str
        self.undub_iso_path: str

        # Create widgets :)
        self.setup_widgets()

        self.undub_thread: threading.Thread

    def setup_widgets(self):
        self.columnconfigure(index=0, weight=0)
        self.columnconfigure(index=1, weight=0)
        self.columnconfigure(index=2, weight=1)

        self.rowconfigure(index=0, weight=0)
        self.rowconfigure(index=1, weight=1)
        self.rowconfigure(index=2, weight=0)

        # ################################
        # Create a Frame for input widgets
        self.widgets_frame = ttk.Frame(self, padding=(0, 0, 0, 0))
        self.widgets_frame.grid(row=0, column=0, padx=(20, 0), pady=(29, 0), sticky="nsew", rowspan=1)
        self.widgets_frame.columnconfigure(index=0, weight=1)

        # Button ISO EU
        self.button_iso_eu = ttk.Button(self.widgets_frame, text="(1) Select European ISO", command=self.select_eu_iso)
        self.button_iso_eu.grid(row=0, column=0, padx=0, pady=0, sticky="nsew", ipadx=10)
        self.add_help(
            self.button_iso_eu,
            """
            Select an untouched European ISO file.
            
            This file will be used as base for the undub.
        """,  # noqa: E501 # pylint: disable=line-too-long
        )

        # Button ISO JP
        self.button_iso_jp = ttk.Button(
            self.widgets_frame, text="(2) Select Japanese ISO", state="disabled", command=self.select_jp_iso
        )
        self.button_iso_jp.grid(row=1, column=0, padx=0, pady=(10, 0), sticky="nsew", ipadx=10)
        self.add_help(
            self.button_iso_jp,
            """
            Select an untouched Japanese ISO file.

            This file will be used to extract japanese voices.
        """,  # noqa: E501 # pylint: disable=line-too-long
        )

        # Button ISO Undub
        self.button_iso_undub = ttk.Button(
            self.widgets_frame, text="(3) Choose Output Undub ISO", state="disabled", command=self.select_undub_iso
        )
        self.button_iso_undub.grid(row=2, column=0, padx=0, pady=(10, 0), sticky="nsew", ipadx=10)
        self.add_help(
            self.button_iso_undub,
            """
            Choose the output path for the undub ISO file.
            
            This will be the path where the undub ISO will be written to.
        """,  # noqa: E501 # pylint: disable=line-too-long
        )

        # ################################
        # Create a Frame for the Hack Checkbuttons
        self.patch_frame = ttk.LabelFrame(self, text="(4) Select Patches", padding=(20, 0))
        self.patch_frame.grid(row=0, column=1, padx=(20, 20), pady=(20, 0), sticky="nsew", rowspan=1)
        self.patch_frame.columnconfigure(index=0, weight=1)

        check_pad_top = 4

        # Checkbuttons
        self.check_fix_kirie_camera_bug = ttk.Checkbutton(
            self.patch_frame, text="Fix Kirie Camera Bug", variable=self.var_fix_kirie_camera_bug, state="disabled"
        )
        self.check_fix_kirie_camera_bug.grid(row=0, column=0, padx=0, pady=(10, check_pad_top), sticky="nsew")
        self.add_help(
            self.check_fix_kirie_camera_bug,
            """
            Fix Kirie Camera Bug.
            
            Fix for the Kirie Camera Bug, where aiming the camera to a certain spot during the last battle with
            Kirie may result in a game freeze.

            Thanks to weirdbeardgame for the fix.
        """,  # noqa: E501 # pylint: disable=line-too-long
        )

        self.check_16_9 = ttk.Checkbutton(self.patch_frame, text="Force 16:9", variable=self.var_16_9, state="disabled")
        self.check_16_9.grid(row=1, column=0, padx=0, pady=check_pad_top, sticky="nsew")
        self.add_help(
            self.check_16_9,
            """
            Enable the 16:9 aspect ratio patch.
            
            Ingame will be rendered in 16:9 and movies will have black bars to preserve the original aspect ratio.
            
            Menus, on screen text and graphics, and the title screen will be stretched.
            
            Remember to force 16:9 aspect ratio in you TV!
        """,  # noqa: E501 # pylint: disable=line-too-long
        )

        self.check_disable_bloom = ttk.Checkbutton(
            self.patch_frame, text="Disable Bloom Effect", variable=self.var_disable_bloom, state="disabled"
        )
        self.check_disable_bloom.grid(row=2, column=0, padx=0, pady=check_pad_top, sticky="nsew")
        self.add_help(
            self.check_disable_bloom,
            """
            Disable ingame bloom effect.
            
            Ingame bloom effect will be disabled.
        """,  # noqa: E501 # pylint: disable=line-too-long
        )

        self.check_remove_dark_filter = ttk.Checkbutton(
            self.patch_frame, text="Remove Dark Filter", variable=self.var_remove_dark_filter, state="disabled"
        )
        self.check_remove_dark_filter.grid(row=3, column=0, padx=0, pady=check_pad_top, sticky="nsew")
        self.add_help(
            self.check_remove_dark_filter,
            """
            Removes ingame dark filter.
            
            Ingame dark filter will be disabled.
        """,  # noqa: E501 # pylint: disable=line-too-long
        )

        self.check_remove_ingame_noise = ttk.Checkbutton(
            self.patch_frame, text="Remove Ingame Noise", variable=self.var_remove_ingame_noise, state="disabled"
        )
        self.check_remove_ingame_noise.grid(row=4, column=0, padx=0, pady=check_pad_top, sticky="nsew")
        self.add_help(
            self.check_remove_ingame_noise,
            """
            Remove ingame noise.
            
            Ingame noise effect will be removed.
        """,  # noqa: E501 # pylint: disable=line-too-long
        )

        self.check_remove_title_noise = ttk.Checkbutton(
            self.patch_frame, text="Remove Title Noise", variable=self.var_remove_title_noise, state="disabled"
        )
        self.check_remove_title_noise.grid(row=5, column=0, padx=0, pady=check_pad_top, sticky="nsew")
        self.add_help(
            self.check_remove_title_noise,
            """
            Remove title noise.
            
            Title noise effect will be removed
        """,  # noqa: E501 # pylint: disable=line-too-long
        )

        self.check_force_language_selection = ttk.Checkbutton(
            self.patch_frame,
            text="Force Language Selection",
            variable=self.var_force_language_selection,
            state="disabled",
        )
        self.check_force_language_selection.grid(row=6, column=0, padx=0, pady=check_pad_top, sticky="nsew")
        self.add_help(
            self.check_force_language_selection,
            """
            Force language selection on launch.
            
            Language selection will be shown at every game launch.
            
            Note that this hack disables loading user preferences at startup completely (as language is one of the preferences).
            This means that settings such as brightness and volume would also be reset to default every time.
        """,  # noqa: E501 # pylint: disable=line-too-long
        )

        self.check_replace_title_jp = ttk.Checkbutton(
            self.patch_frame,
            text="Japanese Title Screen",
            variable=self.var_replace_title_jp,
            state="disabled",
        )
        self.check_replace_title_jp.grid(row=7, column=0, padx=0, pady=check_pad_top, sticky="nsew")
        self.add_help(
            self.check_replace_title_jp,
            """
            Replace title screen with the Japanese one.
            
            The title screen will be replaced with the one found in the Japanese version.
            
            Menu entries and menu fonts will not change as the Japanese ones are not compatible and cannot be replaced.
        """,  # noqa: E501 # pylint: disable=line-too-long
        )

        self.check_replace_models = ttk.Checkbutton(
            self.patch_frame,
            text="Replace Models",
            variable=self.var_replace_models,
            state="disabled",
        )
        self.check_replace_models.grid(row=8, column=0, padx=0, pady=(check_pad_top, 10), sticky="nsew")
        self.add_help(
            self.check_replace_models,
            """
            Replace all Miku models with the original Japanese version.
            
            All Miku models, including both 3D and 2D ones such as the menu portrait and the background image
            displayed at the beginning of each night, will be replaced with the ones found in the Japanese version.
        """,  # noqa: E501 # pylint: disable=line-too-long
        )

        # ################################
        # Progress Frame
        self.progress_frame = ttk.Frame(self, padding=(0, 0, 0, 0))
        self.progress_frame.grid(row=0, column=2, padx=(0, 20), pady=(20, 0), sticky="nsew", rowspan=1, ipadx=0)
        self.progress_frame.columnconfigure(index=0, weight=1)

        # Button Start Undub
        self.button_start_undub = ttk.Button(
            self.progress_frame, text="(5) Start Undub", state="disabled", command=self.start_undub
        )
        self.button_start_undub.grid(row=0, column=0, padx=0, pady=(9, 10), sticky="nsew", ipadx=10)
        self.add_help(
            self.button_start_undub,
            """
            Starts the undub process.
            
            Progress will be shown below by the 6 progress bars. If an error occurs at any point, the corresponding progress bar's text
            will become red, and an error description will be shown here.
        """,  # noqa: E501 # pylint: disable=line-too-long
        )

        # Progressbar
        self.pb_frame_box = ttk.Frame(self.progress_frame, padding=(0, 0, 0, 0))
        self.pb_frame_box.grid(row=1, column=0, padx=(0, 0), pady=(20, 10), sticky="nsew", rowspan=3, ipadx=50)
        self.pb_frame_box.columnconfigure(index=0, weight=0)
        self.pb_frame_box.columnconfigure(index=1, weight=1)
        self.pb_frame_box.columnconfigure(index=2, weight=0, minsize=75)

        def add_pb(_text, i, var, _pady=10):
            label = ttk.Label(self.pb_frame_box, text=_text, state="disabled")
            label.grid(row=i, column=0, pady=(0, _pady), sticky="w")

            progress_undub = ttk.Progressbar(self.pb_frame_box, value=0, variable=var, mode="determinate")
            progress_undub.grid(row=i, column=1, padx=(10, 10), pady=(0, _pady), sticky="ew")

            # Label
            label_progress = ttk.Label(self.pb_frame_box, text="?/?", state="disabled")
            label_progress.grid(row=i, column=2, pady=(0, _pady), sticky="e")

            return label, progress_undub, label_progress

        self.label_eu_check, self.progress_eu_check, self.label_eu_check_perc = add_pb(
            "Checking EU ISO", 0, self.var_progress_eu_check
        )

        self.label_jp_check, self.progress_jp_check, self.label_jp_check_perc = add_pb(
            "Checking JP ISO", 1, self.var_progress_jp_check
        )

        self.label_copy, self.progress_copy, self.label_copy_perc = add_pb("Copying ISO", 2, self.var_progress_copy)

        self.label_movies, self.progress_movies, self.label_movies_perc = add_pb(
            "Patching Movies", 3, self.var_progress_movies
        )

        self.label_elf, self.progress_elf, self.label_elf_perc = add_pb("Patching ELF", 4, self.var_progress_elf)

        self.label_contents, self.progress_contents, self.label_contents_perc = add_pb(
            "Patching Contents", 5, self.var_progress_contents
        )

        # Button Done
        self.button_done = ttk.Button(self.progress_frame, text="Done", state="disabled", command=self.undub_done)
        self.button_done.grid(row=6, column=0, padx=0, pady=(9, 10), sticky="nsew", ipadx=10)

        # ################################
        self.info_frame = ttk.LabelFrame(self, text="Info", padding=(20, 0))
        self.info_frame.grid(row=1, column=0, padx=(20, 20), pady=(0, 0), sticky="nsew", rowspan=1, columnspan=3)
        self.info_frame.columnconfigure(index=0, weight=1)
        self.info_frame.rowconfigure(index=0, weight=1, minsize=200)

        _default_text = format_help_text(
            """
            Welcome to karas84's Project Zero Undubber!
            
            Follow the 5 steps to create your custom undub. You will need:
            
            $ - An original and untouched European ISO of Project Zero (SLES 508.21)
            $ - An original and untouched Japanese ISO of Zero (SLPS 250.74)
            
            If you like, you can choose to apply extra patches such as the 16:9 aspect ratio or removing the ingame bloom effect.
            
            You can obtain more help by hovering the mouse on each component (e.g., a button or a checkbox) of this window.
            You'll see here the help for that specific component.
        """  # noqa: E501 # pylint: disable=line-too-long
        )
        self.label_info = ttk.Label(self.info_frame, text=_default_text)
        self.label_info.configure(anchor="nw")
        self.label_info.grid(row=0, column=0, pady=(0, 10), sticky="nsew")
        setattr(self.label_info, "_default_text", _default_text)

        # ################################
        # Sizegrip
        self.sizegrip = ttk.Sizegrip(self)
        self.sizegrip.grid(row=3, column=2, padx=(0, 5), pady=(0, 5), sticky="e")

    def select_eu_iso(self):
        eu_iso_path = fd.askopenfilename(
            title="Select European ISO",
            initialdir=os.path.expanduser("~"),
            filetypes=self.filetypes,
        )

        if eu_iso_path:
            self.eu_iso_path = eu_iso_path
            self.button_iso_jp.state(["!disabled"])

    def select_jp_iso(self):
        jp_iso_path = fd.askopenfilename(
            title="Select Japanese ISO",
            initialdir=os.path.expanduser("~"),
            filetypes=self.filetypes,
        )

        if jp_iso_path:
            self.jp_iso_path = jp_iso_path
            self.button_iso_undub.state(["!disabled"])

    def select_undub_iso(self):
        undub_iso_path = fd.asksaveasfilename(
            title="Choose Output Undub ISO",
            initialdir=os.path.expanduser("~"),
            filetypes=self.filetypes,
        )

        if undub_iso_path:
            if os.path.splitext(undub_iso_path.lower())[1] != ".iso":
                undub_iso_path += ".iso"
            self.undub_iso_path = undub_iso_path
            self.check_fix_kirie_camera_bug.state(["!disabled"])
            self.check_16_9.state(["!disabled"])
            self.check_disable_bloom.state(["!disabled"])
            self.check_remove_dark_filter.state(["!disabled"])
            self.check_remove_ingame_noise.state(["!disabled"])
            self.check_remove_title_noise.state(["!disabled"])
            self.check_force_language_selection.state(["!disabled"])
            self.check_replace_title_jp.state(["!disabled"])
            self.check_replace_models.state(["!disabled"])
            self.button_start_undub.state(["!disabled"])

    def start_undub(self):
        self.button_iso_eu.state(["disabled"])
        self.button_iso_jp.state(["disabled"])
        self.button_iso_undub.state(["disabled"])
        self.check_fix_kirie_camera_bug.state(["disabled"])
        self.check_16_9.state(["disabled"])
        self.check_disable_bloom.state(["disabled"])
        self.check_remove_dark_filter.state(["disabled"])
        self.check_remove_ingame_noise.state(["disabled"])
        self.check_remove_title_noise.state(["disabled"])
        self.check_force_language_selection.state(["disabled"])
        self.check_replace_title_jp.state(["disabled"])
        self.check_replace_models.state(["disabled"])
        self.button_start_undub.state(["disabled"])

        self._undub_started = True
        self.label_info["text"] = format_help_text(
            """
            Undubbing process started.
            
            Please wait while the game is being patched.
        """
        )

        self.undub_thread = UndubThread(self)
        self.undub_thread.daemon = True
        self.undub_thread.start()

    def undub_done(self):
        self.quit()

    def add_help(self, widget, help_text):
        setattr(widget, "_help_text", format_help_text(help_text))
        widget.bind("<Enter>", self._show_help)
        widget.bind("<Leave>", self._show_help)

    def _show_help(self, event: tk.Event):
        if self._undub_started:
            return

        text = getattr(self.label_info, "_default_text")
        if event.type == tk.EventType.Enter:
            text = getattr(event.widget, "_help_text", text)
        # elif event.type == tk.EventType.Leave:
        self.label_info["text"] = text

    @staticmethod
    def _make_red(widget):
        widget.configure(foreground="#E30B5C")


def main():
    parser = argparse.ArgumentParser(description="Project Zero Undub GUI")

    parser.add_argument(
        "-f",
        "--foreground",
        dest="foreground",
        action="store_true",
        help="do not detach from shell (unix only; on windows always run in foreground)",
    )

    args = parser.parse_args()

    if not args.foreground and hasattr(os, "fork") and os.fork():
        sys.exit()

    root = tk.Tk()
    root.title("Project Zero Undub by karas84")

    style = ttk.Style(root)
    style.theme_use("default")

    app = App(root)
    app.pack(fill="both", expand=True)

    # Set a minsize for the window, and place it in the middle
    root.update()
    root.minsize(root.winfo_width(), root.winfo_height())
    x_coordinate = int((root.winfo_screenwidth() / 2) - (root.winfo_width() / 2))
    y_coordinate = int((root.winfo_screenheight() / 2) - (root.winfo_height() / 2))
    root.geometry(f"+{x_coordinate}+{y_coordinate - 20}")

    root.mainloop()



from contextlib import contextmanager

import pycdlib


class CDDVD:
    def __init__(self, iso_path: str):
        self.iso = pycdlib.PyCdlib()
        self.iso.open(iso_path, mode="rb")
        self.iso9660_facade = self.iso.get_iso9660_facade()

        # name to iso name (e.g. README.TXT -> README.TXT;1)
        self.files = {file_name.rsplit(";", 1)[0]: file_name for file_name in next(self.iso9660_facade.walk("/"))[2]}

    def find_file(self, file_name: str):
        upper_name = file_name.upper()

        if upper_name in self.files:
            file_name = self.files[upper_name]

        elif file_name not in self.files.values():
            return None

        return file_name

    def read_file(self, file_name: str, size=-1, offset=0):
        try:
            with self.open(file_name) as file_h:
                file_h.seek(offset)
                return file_h.read(size)
        except FileNotFoundError:
            return None

    @contextmanager
    def open(self, file_name: str):
        _file_name = self.find_file(file_name)

        if not _file_name:
            raise FileNotFoundError(file_name)

        with self.iso9660_facade.open_file_from_iso(f"/{_file_name}") as file_h:
            yield file_h

    def get_file_size(self, file_name):
        file_name = self.find_file(file_name)

        if file_name is None:
            return None

        return self.iso9660_facade.get_record(f"/{file_name}").get_data_length()

    def __delete__(self, instance):
        self.iso.close()



# This file is copyright of William Adam-Grenier and is released
# under the GNU General Public License v3.0. The following is a
# slightly modified version of the original, which can be found
# in the author's github repository: https://github.com/wagrenier/PssMux

import os
import io
import glob

from shutil import copyfile
from typing import BinaryIO


audio_segment = b"\x00\x00\x01\xBD"
pack_start = b"\x00\x00\x01\xBA"
end_file = b"\x00\x00\x01\xB9"


first_header_size = 0x3F
header_size = 0x17


def seek_next_audio(file):
    pos = file.tell()
    file.seek(0, os.SEEK_SET)
    sz = file.seek(0, os.SEEK_END)
    file.seek(pos)

    while True:
        block_id = file.read(0x4)

        if block_id == pack_start:
            file.seek(0xA, os.SEEK_CUR)
        elif block_id == audio_segment:
            return False
        elif block_id == end_file:
            return True
        elif file.tell() >= sz:
            return True
        else:
            block_size = file.read(0x2)
            file.seek(int.from_bytes(block_size, "big"), os.SEEK_CUR)


def initial_audio_block(file):
    b_size = int.from_bytes(file.read(0x2), "big")

    file.seek(0x3B - 0x6, os.SEEK_CUR)

    audio_total_size = int.from_bytes(file.read(0x4), "little")
    data_size = b_size - first_header_size + 0x6

    return audio_total_size, data_size


def audio_block(file):
    b_size = int.from_bytes(file.read(0x2), "big")

    file.seek(-0x6, os.SEEK_CUR)
    file.seek(header_size, os.SEEK_CUR)

    data_size = b_size - header_size + 0x6

    return data_size


def build_full_audio_buffer_io(file):
    seek_next_audio(file)
    total_size, curr_block_size = initial_audio_block(file)
    buff = io.BytesIO()
    buff.write(file.read(curr_block_size))

    while True:
        if seek_next_audio(file):
            break

        curr_block_size = audio_block(file)
        buff.write(file.read(curr_block_size))

    file.seek(0)
    buff.seek(0)
    return buff


def pss_mux(source: str, target: str, output: str):
    copyfile(target, output)

    with open(source, "rb") as source_file:
        with open(output, "rb+") as target_file:
            pss_mux_from_bytes_io(source_file, target_file)


def pss_mux_inplace(source: str, target: str):
    with open(source, "rb") as source_file:
        with open(target, "rb+") as target_file:
            pss_mux_from_bytes_io(source_file, target_file)


def pss_mux_in_memory(source: str, target: str):
    with open(source, "rb") as source_file:
        with open(target, "rb") as target_file:
            target_buffer_io = io.BytesIO(target_file.read())

            pss_mux_from_bytes_io(source_file, target_buffer_io)

            target_buffer_io.seek(0)
            return target_buffer_io


def pss_mux_from_bytes_io(source_io: BinaryIO, target_io: BinaryIO):
    total_buffer_written = 0x0
    source_full_buff_io = build_full_audio_buffer_io(source_io)

    seek_next_audio(target_io)
    seek_next_audio(source_io)

    target_total_size, target_curr_block_size = initial_audio_block(target_io)
    source_total_size, source_curr_block_size = initial_audio_block(source_io)

    source_io.seek(source_curr_block_size, os.SEEK_CUR)

    data = source_full_buff_io.read(target_curr_block_size)
    target_io.write(data)

    total_buffer_written += target_curr_block_size

    while True:
        target_done = seek_next_audio(target_io)
        source_done = seek_next_audio(source_io)

        if target_done or source_done:
            break

        target_curr_block_size = audio_block(target_io)
        source_curr_block_size = audio_block(source_io)

        source_io.seek(source_curr_block_size, os.SEEK_CUR)

        data = source_full_buff_io.read(target_curr_block_size)
        target_io.write(data)

        total_buffer_written += target_curr_block_size

    return_value = target_io.read()

    return return_value


def parse_all_videos(jp_path, en_path, out_path):
    jp_video_movie = [f for f in glob.glob(os.path.join(jp_path, "MOVIE", "*.*")) if f.upper().endswith(".PSS")]
    jp_video_movie2 = [f for f in glob.glob(os.path.join(jp_path, "MOVIE2", "*.*")) if f.upper().endswith(".PSS")]

    en_video_movie = [f for f in glob.glob(os.path.join(en_path, "MOVIE", "*.*")) if f.upper().endswith(".PSS")]
    en_video_movie2 = [f for f in glob.glob(os.path.join(en_path, "MOVIE2", "*.*")) if f.upper().endswith(".PSS")]
    en_video_movie3 = [f for f in glob.glob(os.path.join(en_path, "MOVIE3", "*.*")) if f.upper().endswith(".PSS")]
    en_video_movie4 = [f for f in glob.glob(os.path.join(en_path, "MOVIE4", "*.*")) if f.upper().endswith(".PSS")]

    en_video_movie5 = [f for f in glob.glob(os.path.join(en_path, "MOVIE5", "*.*")) if f.upper().endswith(".PSS")]

    def find_matches(_jp_list, _en_list):
        def basename(p):
            return os.path.splitext(os.path.basename(p))[0].rstrip("p")

        matches = list()
        for _jp_movie in _jp_list:
            en_match = next((_en_movie for _en_movie in _en_list if basename(_en_movie) == basename(_jp_movie)), None)
            if en_match:
                matches.append((_jp_movie, en_match))

        return matches

    matches_movie = find_matches(jp_video_movie, en_video_movie)
    matches_movie_p = find_matches(jp_video_movie, en_video_movie3)

    matches_movie2 = find_matches(jp_video_movie2, en_video_movie2)
    matches_movie2_p = find_matches(jp_video_movie2, en_video_movie4)

    def basedir(p):
        return os.path.basename(os.path.dirname(p))

    def st_size(f):
        return os.stat(f).st_size

    for jp_movie, en_movie in matches_movie + matches_movie_p + matches_movie2 + matches_movie2_p:
        out_movie = os.path.join(out_path, basedir(en_movie), os.path.basename(en_movie))
        os.makedirs(os.path.dirname(out_movie), exist_ok=True)
        pss_mux(jp_movie, en_movie, out_movie)
        print(out_movie)
        assert st_size(en_movie) == st_size(out_movie)

    for en_movie in en_video_movie5:
        out_movie = os.path.join(out_path, basedir(en_movie), os.path.basename(en_movie))
        os.makedirs(os.path.dirname(out_movie), exist_ok=True)
        print(out_movie)
        copyfile(en_movie, out_movie)



import os
import math


def copy_file(src, dst, callback=None):
    with open(src, "rb") as src_fh:
        src_size = src_fh.seek(0, os.SEEK_END)
        src_fh.seek(0, os.SEEK_SET)

        block_size = 2**20

        if callback:
            block_num = int(math.ceil(src_size / block_size))
            callback(block_num)

        with open(dst, "wb") as dst_fh:
            while True:
                data = src_fh.read(block_size)

                if not data:
                    break

                dst_fh.write(data)

                if callback:
                    callback()



import os
import sys
import argparse

from dataclasses import dataclass

from ..error import AbstractProgressError
from ..progress import Progress
from ..utils import copy_file
from ...zero.iso import (
    check_iso_hashes,
    replace_movies_in_iso_inplace,
    patch_elf_inplace,
    merge_iso_img_bd_contents,
)


@dataclass
class ProgressErrorFileExist(AbstractProgressError):
    error_msg = "Error while accessing ISO files"
    explanation = "Some input ISO files cannot be accessed or read"
    suggestion = "Please check that the provided files exist and that they can be read"


@dataclass
class ProgressErrorCRCEU(AbstractProgressError):
    error_msg = "Error while checking EU ISO files"
    explanation = "Some files are not how they are supposed to be. This may be due to a bad dump"
    suggestion = "Please re-dump the game and try again"


@dataclass
class ProgressErrorCRCJP(AbstractProgressError):
    error_msg = "Error while checking JP ISO files"
    explanation = "Some files are not how they are supposed to be. This may be due to a bad dump"
    suggestion = "Please re-dump the game and try again"


@dataclass
class ProgressErrorISOCopy(AbstractProgressError):
    error_msg = "Error while copying ISO file"
    explanation = "Disk may be full or you may be trying to write somewhere you are not allowed to"
    suggestion = "Check disk space and destination folder and try again"


@dataclass
class ProgressErrorMoviePatch(AbstractProgressError):
    error_msg = "Error while patching Movies"
    explanation = "This is not supposed to happen if previous ISO checks were completed successfully"
    suggestion = "Please report this issue on the GitHub page"


@dataclass
class ProgressErrorELFPatch(AbstractProgressError):
    error_msg = "Error while patching ELF file"
    explanation = "This is not supposed to happen if previous ISO checks were completed successfully"
    suggestion = "Please report this issue on the GitHub page"


@dataclass
class ProgressErrorGamePatch(AbstractProgressError):
    error_msg = "Error while patching game contents"
    explanation = "This is not supposed to happen if previous ISO checks were completed successfully"
    suggestion = "Please report this issue on the GitHub page"


def main():
    parser = argparse.ArgumentParser(description="Project Zero Undub Tool")

    parser.add_argument("iso_eu", type=str, help="path to the European ISO (SLES508.21)")
    parser.add_argument("iso_jp", type=str, help="path to the Japanese ISO (SLPS250.74)")
    parser.add_argument("iso_out", type=str, help="output path for the undubbed ISO (file MUST NOT exist")

    parser.add_argument(
        "--fix-kirie-camera-bug",
        dest="fix_kirie_camera_bug",
        action="store_true",
        help="Fixes a bug where the game could freeze during the last battle with Kirie",
    )
    parser.add_argument(
        "--jp-title",
        dest="replace_title_jp",
        action="store_true",
        help="Replace the title screen background with the one in the japanese version",
    )
    parser.add_argument(
        "--replace-models",
        dest="replace_models",
        action="store_true",
        help="Replace all Miku models with the original Japanese version",
    )
    parser.add_argument(
        "--force-language-selection",
        dest="force_lang",
        action="store_true",
        help="Force the language selection menu to appear every time",
    )
    parser.add_argument("--disable-bloom", dest="no_bloom", action="store_true", help="Remove in-game bloom effect")
    parser.add_argument(
        "--remove-dark-filter",
        dest="dark_filter",
        action="store_true",
        help="Remove in-game dark filter (makes the game slightly brighter)",
    )
    parser.add_argument(
        "--remove-ingame-noise", dest="ingame_noise", action="store_true", help="Remove in-game noise effect"
    )
    parser.add_argument(
        "--remove-title-noise", dest="menu_noise", action="store_true", help="Remove noise effect in title menu"
    )
    parser.add_argument(
        "--force-16-9-ingame",
        dest="force_16_9_game",
        action="store_true",
        help="Force game to render graphics in 16:9 format",
    )
    parser.add_argument(
        "--force-16-9-movies",
        dest="force_16_9_movies",
        action="store_true",
        help="Force movies to render in 16:9 format (to be used in conjunction with the 16:9 "
        "in-game flag to add black bars around movies, otherwise they will be stretched)",
    )

    args = parser.parse_args()

    if os.path.exists(args.iso_out):
        print()
        print("  Output ISO already exists. This program will not overwrite any")
        print("  existing file. You should specify a new path for the output.")
        print()
        sys.exit(1)

    args.force_16_9_movies &= args.force_16_9_game

    if not os.path.exists(args.iso_eu) or not os.path.exists(args.iso_jp):
        ProgressErrorFileExist.print()
        sys.exit(1)

    print()
    print("  Welcome to the Project Zero undubbing process (by karas84)!")
    print("  Please wait while your game is being undubbed...")
    print()

    with Progress("Checking EU ISO", ProgressErrorCRCEU) as pbar:
        check_iso_hashes(args.iso_eu, lang="EU", callback=pbar)

    with Progress("Checking JP ISO", ProgressErrorCRCJP) as pbar:
        check_iso_hashes(args.iso_jp, lang="JP", callback=pbar)

    with Progress("Copying ISO", ProgressErrorISOCopy) as pbar:
        copy_file(args.iso_eu, args.iso_out, pbar)

    with Progress("Patching Movies", ProgressErrorMoviePatch) as pbar:
        replace_movies_in_iso_inplace(args.iso_jp, args.iso_out, pbar)

    with Progress("Patching ELF", ProgressErrorELFPatch) as pbar:
        patch_elf_inplace(
            args.iso_out,
            fix_kirie_camera_bug=args.fix_kirie_camera_bug,
            force_lang=args.force_lang,
            no_bloom=args.no_bloom,
            dark_filter=args.dark_filter,
            ingame_noise=args.ingame_noise,
            menu_noise=args.menu_noise,
            force_16_9_game=args.force_16_9_game,
            force_16_9_movies=args.force_16_9_movies,
            callback=pbar,
        )

    with Progress("Patching Contents", ProgressErrorGamePatch) as pbar:
        merge_iso_img_bd_contents(
            eu_iso_path=args.iso_eu,
            jp_iso_path=args.iso_jp,
            out_iso_path=args.iso_out,
            replace_title_jp=args.replace_title_jp,
            replace_models=args.replace_models,
            replace_sfx=True,
            callback=pbar,
        )

    print()
    print("  All done. Enjoy the game!")
    print()


if __name__ == "__main__":
    main()



from .utils import make_table


_font_01 = """
　ＡＢＣＤＥＦＧＨＩＪＫＬＭＮＯＰＱＲＳＴ
ＵＶＷＸＹＺａｂｃｄｅｆｇｈｉｊｋｌｍｎｏ
ｐｑｒｓｔｕｖｗｘｙｚ０１２３４５６７８９
{０}{１}{２}{３}{４}{５}{６}{７}{８}{９}あいうえおかきくけこさ
しすせそたちつてとなにぬねのはひふへほまみ
むめもやゆよらりるれろわをんアイウエオカキ
クケコサシスセソタチツテトナニヌネノハヒフ
ヘホマミムメモヤユヨラリルレロワヲンぁ{○}{✕}
{△}{□}ゃゅょっァィゥェォャュョッがぎぐげござ
じずぜぞだ終づでどばびぶべぼぱぴ序充攻ガギ
"""


_font_02 = """
グゲゴザジズゼゾダ溜貯デドバビブベボパピプ
ぺポヴ（）ー？/、{、}。：{、2}{、3}！「」✓✗{pts}　
零一二三四五六七八九十百千万上下前後左右扉
固閉咮面掛屏風向编集霊取材手入階段動棚冬雑
然人形並止押中何不気者廊犠影輿消鏡見返振美
琴音閒夭井粱兄…写鳥居古牲常敗化除効果低用
感度良高暗遠撮钵力少回復薬葉書容器參精神全
香料強肆静作持位槽清水異{常}黑石付伍带性壊電
灯方闇照出母深雪遺思議械目映鍵地銀製小飾赤
鲭塚式陸丸月漆御捌玫射機＝範纳戸足欠仏像金
"""


_font_03 = """
曲懐燈大青銅墨文字木礼型再生部分破崩彫底梵
浮〇台追屋敷内圧襖無着物明拾軸今開助隙間障
子女斎真救花使呪規家姿撃氷室秘現封印解関記
録光陣緑側穴覚予桜走拡供引装填必要状態邸玄
倉紋院瓦礫庭新館通路眺望夜完峰湧元逢族娘長
髪保存伝承他切替削護在篭淵衝立臨縄和鳴白鎧
時計奈反箪笥行控導錠指置寂打外暖吹込急縛男
溶倒進死殺苦魂憎繋厨号病証亡痛続最初配戻離
選年日午秒枚数退官得点正操術客能調整画牢落
巨半口探格布若連想惹原留床空裂冷流包里悟跡
"""


_font_04 = """
残誰突嫌近螨燭本工事儀柱囲壁散暴定自視根血
骨紫董住趣相当偭値知特表纸仁王炉裏積針巣張
語渡泣描绘簿備变庫橋錡麗皿窓姐色悲々火鉢頑
丈机割霧淼迈景湿含瓶毛同意奇妙塗座代弐片隅
傷箱草吊乱役邪魔来奉痦卷牌毬移土砂武周注穢
蒼伸顏塞碑慰{色}枝壱堂禊先端抜厚竹墓読設黄応
更声量節確認平坂絡有次陽品欲許碧参加準星巴
緒浩私名喜哀補的図基隣獅民車場所話荒言資鍮
象板譜悪章謝道迷示唱報等査広収主別念述員液
傪命绝妻廃墟红理由杭恐俗学各著宗蔵依利境重
"""


_font_05 = """
心失踪报好昨遅腕宿速殊会震村起説題樣択我以
増背帳断潰戦件仕未詳细鎖避対巫達越研究成世
去発実多隠違治執殿建禁折彼凶劇熱憶狂信氏両
首肢昔係驚早岩案角待造呼考赞害過運逃夫既致
法似合情災厄怪界門因社般監鐘測豊富结論忌推
療養葬始直了彩頼胸騷泉楽{彩}限歩鬼幼差訴帰植
怒怖遊誘迫麓幻策{Rune4}碎寒夢教識頭捕謎安{Rune1}{Rune2}酷
怨晴第舞県山{Rune5}祭吉兆南胴警察{Rune3}故捜身歲野児
童届噂{策}師衰弱質問親翌決雛咲壇息友弔守司筋
極虐放協瘴禍刻襲蘇降際登構宮稿〜国仰夹崖蓄
"""


table_jp = {
    "default": make_table(_font_01),
    0xF0: make_table(_font_02),
    0xF1: make_table(_font_03),
    0xF2: make_table(_font_04),
    0xF3: make_table(_font_05),
}



import os
import re
import struct

from typing import BinaryIO, cast
from contextlib import contextmanager

from .entry import TOCEntry, FileEntry
from .adapter import FilesystemAdapter, ISOAdapter
from ...utils.file import SubFile


def is_iso_file(file_path):
    return os.path.isfile(file_path) and os.path.splitext(file_path)[1].upper() == ".ISO"


class PJZReader:
    re_dat_entry = re.compile(r"([A-Z0-9_]+)_([A-Z0-9]+):([0-9]+)")

    def __init__(self, load_path):
        self.load_path = load_path

        if is_iso_file(load_path):
            self.adapter = ISOAdapter(self.load_path)
        elif os.path.isdir(load_path):
            self.adapter = FilesystemAdapter(self.load_path)
        else:
            raise ValueError("load_path must be iso or filesystem path")

        self.adapter.setup()

        toc_entry_list = self.prepare_toc_and_entries()

        self.toc_entry_list = toc_entry_list

        self.file_name_index = self.make_file_name_index()

    def parse_cd_file_dat(self, elf_path):
        elf_bin = self.adapter.read_file(elf_path)

        if not elf_bin:
            return None

        try:
            cd_file_dat_start = elf_bin.index(b"CD_FILE_DAT:T")
            cd_file_dat_end = elf_bin.index(b",;", cd_file_dat_start)
            cd_file_dat = elf_bin[cd_file_dat_start:cd_file_dat_end]
        except ValueError:
            return None

        cd_file_dat = cd_file_dat.replace(b"\x2C\x5C\x00", b"\x2C").decode("ascii")
        section_name, file_list = cd_file_dat.split("=e", 1)

        file_list = file_list.rstrip(",").split(",")

        entry_matches = [self.re_dat_entry.match(entry) for entry in file_list]

        if not all(entry_matches):
            return None

        entry_matches = cast(list[re.Match[str]], entry_matches)

        toc_entry_list = [
            TOCEntry(f"{entry.group(1)}.{entry.group(2)}", int(entry.group(3))) for entry in entry_matches
        ]

        return toc_entry_list

    def parse_img_hd(self, img_hd_path):
        img_hd_bin = self.adapter.read_file(img_hd_path)

        if not img_hd_bin:
            return None

        if len(img_hd_bin) % 8 != 0:
            return None

        num_int32 = len(img_hd_bin) // 4
        file_data = struct.unpack(f"<{num_int32}I", img_hd_bin)

        file_entry_list = [FileEntry(file_data[i] * 0x800, file_data[i + 1]) for i in range(0, len(file_data), 2)]

        return file_entry_list

    @staticmethod
    def validate_toc(toc_entry_list, file_entry_list):
        return (
            toc_entry_list is not None and file_entry_list is not None and len(toc_entry_list) == len(file_entry_list)
        )

    def validate_image_bd_size(self, file_entry_list):
        max_file_entry_position = max([fe.offset + fe.size for fe in file_entry_list])
        image_bd_size = self.adapter.get_img_bd_size()
        return image_bd_size >= max_file_entry_position

    def print_files(self):
        for toc_entry in self.toc_entry_list:
            print(f"{toc_entry.name} {toc_entry.size}")

    def list_files(self) -> list[str]:
        return [toc_entry.name for toc_entry in self.toc_entry_list]

    def num_files(self):
        return len(self.toc_entry_list)

    def file_exist(self, file_name):
        return file_name in ("", ".") or file_name in self.file_name_index

    def find_entry(self, file_name) -> "TOCEntry | None":
        if file_name not in self.file_name_index:
            return None

        toc_entry = self.file_name_index[file_name]

        return toc_entry

    def read_file(self, file_name, size=-1, offset=0):
        assert self.adapter.img_bd_path is not None

        toc_entry = self.find_entry(file_name)

        if toc_entry is None:
            return RuntimeError("file not found")

        bytes_left = max(0, toc_entry.size - offset)
        if size == -1 or size > bytes_left:
            size = bytes_left

        return self.adapter.read_file(self.adapter.img_bd_path, size, toc_entry.offset + offset)

    @contextmanager
    def open(self, file_name):
        assert self.adapter.img_bd_path is not None

        toc_entry = self.find_entry(file_name)

        if toc_entry is None:
            return RuntimeError("file not found")

        with self.adapter.open(self.adapter.img_bd_path) as file_h:
            file_h = cast(BinaryIO, file_h)
            yield SubFile(file_h, toc_entry.offset, toc_entry.size)  # pylint: disable=abstract-class-instantiated

    def prepare_toc_and_entries(self):
        if not (self.adapter.elf_path and self.adapter.img_hd_path and self.adapter.img_bd_path):
            raise RuntimeError("missing adapter paths for elf or img_hd or img_bd")

        if not all(
            map(self.adapter.test_file, (self.adapter.elf_path, self.adapter.img_hd_path, self.adapter.img_bd_path))
        ):
            raise RuntimeError("wrong root folder")

        toc_entry_list = self.parse_cd_file_dat(self.adapter.elf_path)
        file_entry_list = self.parse_img_hd(self.adapter.img_hd_path)

        if not toc_entry_list or not file_entry_list:
            raise RuntimeError("cannot read elf for entries")

        if not self.validate_toc(toc_entry_list, file_entry_list) or not self.validate_image_bd_size(file_entry_list):
            raise RuntimeError("wrong format")

        for toc_entry, file_entry in zip(toc_entry_list, file_entry_list):
            toc_entry.offset = file_entry.offset
            toc_entry.size = file_entry.size

        return toc_entry_list

    def make_file_name_index(self) -> dict[str, TOCEntry]:
        return {entry.name: entry for entry in self.toc_entry_list}

    def __repr__(self):
        return (
            f"{self.__class__.__name__}[{self.adapter.__class__.__name__}="
            f"{os.path.basename(self.adapter.load_path)}]"
        )



from io import BytesIO
from abc import ABC, abstractmethod
from contextlib import contextmanager


class AbstractAdapter(ABC):
    @abstractmethod
    def test_file(self, file_path: str) -> bool:
        ...

    @abstractmethod
    def read_file(self, file_name: str, size=-1, offset=0) -> bytes:
        ...

    @abstractmethod
    def get_img_bd_size(self) -> int:
        ...

    @abstractmethod
    def find_file(self, file_name: str) -> str:
        ...

    @contextmanager
    @abstractmethod
    def open(self, file_name: str) -> BytesIO:
        ...

    def __init__(self, load_path: str):
        self._init_called = True

        self.load_path = load_path

        self.elf_path = None
        self.img_hd_path = None
        self.img_bd_path = None

    def setup(self):
        if not getattr(self, "_init_called", False):
            raise RuntimeError("must call __init__ of super class")

        self.elf_path = self.find_file("SLES_508.21") or self.find_file("SLPS_250.74")
        self.img_hd_path = self.find_file("IMG_HD.BIN")
        self.img_bd_path = self.find_file("IMG_BD.BIN")
